import os
import unittest
import warnings  # noqa

import numpy as np
import pandas as pd
from sklearn.compose import ColumnTransformer
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.impute import SimpleImputer
from sklearn.metrics import make_scorer, f1_score, precision_score, recall_score, SCORERS, roc_auc_score, \
    fbeta_score, cohen_kappa_score, \
    confusion_matrix, mean_squared_error, mean_absolute_error
from sklearn.model_selection import train_test_split, GridSearchCV, RepeatedKFold
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import label_binarize, StandardScaler, OneHotEncoder

import helpsk as hlp
from helpsk.exceptions import HelpskAssertionError
from helpsk.pandas import print_dataframe
from helpsk.sklearn_eval import MLExperimentResults, TwoClassEvaluator, RegressionEvaluator
from helpsk.sklearn_pipeline import CustomOrdinalEncoder
from helpsk.utility import redirect_stdout_to_file
from tests.helpers import get_data_credit, get_test_path, check_plot, helper_test_dataframe, \
    get_data_housing, clean_formatted_dataframe


def warn(*args, **kwargs):  # noqa
    pass
warnings.warn = warn  # noqa


# noinspection PyMethodMayBeStatic
class TestSklearnEval(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        ####
        # set up grid-search on classification model for credit data
        ####

        credit_data = get_data_credit()
        credit_data.loc[0:46, ['duration']] = np.nan
        credit_data.loc[25:75, ['checking_status']] = np.nan
        credit_data.loc[10:54, ['credit_amount']] = 0
        y_full = credit_data['target']
        X_full = credit_data.drop(columns='target')  # noqa
        y_full = label_binarize(y_full, classes=['good', 'bad']).flatten()
        X_train, X_test, y_train, y_test = train_test_split(X_full, y_full, test_size=0.2, random_state=42)  # noqa
        del y_full, X_full
        numeric_columns = hlp.pandas.get_numeric_columns(X_train)
        non_numeric_columns = hlp.pandas.get_non_numeric_columns(X_train)
        numeric_pipeline = Pipeline([
            ('imputing', SimpleImputer(strategy='mean')),
            ('scaling', StandardScaler()),
        ])
        non_numeric_pipeline = Pipeline([
            ('encoder_chooser', hlp.sklearn_pipeline.TransformerChooser()),
        ])
        transformations_pipeline = ColumnTransformer([
            ('numeric_pipeline', numeric_pipeline, numeric_columns),
            ('non_numeric_pipeline', non_numeric_pipeline, non_numeric_columns)
        ])
        random_forest_model = RandomForestClassifier(random_state=42)
        full_pipeline = Pipeline([
            ('preparation', transformations_pipeline),
            ('model', random_forest_model)
        ])
        param_grad = [
            {
                'preparation__non_numeric_pipeline__encoder_chooser__transformer': [OneHotEncoder(),
                                                                                    CustomOrdinalEncoder()],
                'model__min_samples_split': [2],  # test zero-variance params
                'model__max_features': [100, 'auto'],
                'model__n_estimators': [10, 50],
            },
        ]
        # https://github.com/scikit-learn/scikit-learn/blob/2beed55847ee70d363bdbfe14ee4401438fba057/sklearn_eval/metrics/_scorer.py#L702
        # https://scikit-learn.org/stable/modules/generated/sklearn.metrics.make_scorer.html
        # https://stackoverflow.com/questions/60615281/different-result-roc-auc-score-and-plot-roc-curve
        scores = {
            # https://github.com/scikit-learn/scikit-learn/blob/2beed5584/sklearn_eval/metrics/_scorer.py#L537
            'ROC/AUC': SCORERS['roc_auc'],
            'F1': make_scorer(f1_score, greater_is_better=True),
            'Pos. Pred. Val': make_scorer(precision_score, greater_is_better=True),
            'True Pos. Rate': make_scorer(recall_score, greater_is_better=True),
        }
        grid_search = GridSearchCV(full_pipeline,
                                   param_grid=param_grad,
                                   cv=RepeatedKFold(n_splits=3, n_repeats=1, random_state=42),
                                   scoring=scores,
                                   refit=False,
                                   return_train_score=True)
        grid_search.fit(X_train, y_train)
        cls.credit_data__grid_search = grid_search

        param_grad = [
            {
                'preparation__non_numeric_pipeline__encoder_chooser__transformer': [OneHotEncoder(),
                                                                                    CustomOrdinalEncoder()],
                'model__max_features': [100, 'auto'],
                'model__n_estimators': [10, 50],
            },
        ]
        grid_search = GridSearchCV(full_pipeline,
                                   param_grid=param_grad,
                                   cv=RepeatedKFold(n_splits=3, n_repeats=1, random_state=42),
                                   scoring='roc_auc',
                                   refit=True,
                                   return_train_score=False)
        grid_search.fit(X_train, y_train)
        cls.credit_data__grid_search__roc_auc = grid_search

        best_model = grid_search.best_estimator_
        predicted_scores = best_model.predict_proba(X_test)[:, 1]
        cls.credit_data__y_test = y_test
        cls.credit_data__y_scores = predicted_scores

        ####
        # set up grid-search on regression model for housing data
        ####
        housing_data = get_data_housing()
        housing_data.loc[0:46, ['median_income']] = np.nan
        housing_data.loc[25:75, ['housing_median_age']] = np.nan
        y_full = housing_data['target']
        X_full = housing_data.drop(columns='target')  # noqa
        X_train, X_test, y_train, y_test = train_test_split(X_full, y_full, test_size=0.2, random_state=42)  # noqa
        del y_full, X_full
        numeric_pipeline = Pipeline([
            ('imputing', SimpleImputer(strategy='mean')),
            ('scaling', StandardScaler()),
        ])
        random_forest_model = RandomForestRegressor(random_state=42)
        full_pipeline = Pipeline([
            ('preparation', numeric_pipeline),
            ('model', random_forest_model)
        ])
        param_grad = [
            {
                'model__max_features': [2, 'auto'],
                'model__n_estimators': [10, 50]
            },
        ]
        scores = {
            'RMSE': make_scorer(mean_squared_error, greater_is_better=False, squared=False),
            'MAE': make_scorer(mean_absolute_error, greater_is_better=False),
        }
        grid_search = GridSearchCV(full_pipeline,
                                   param_grid=param_grad,
                                   cv=RepeatedKFold(n_splits=3, n_repeats=1, random_state=42),
                                   scoring=scores,
                                   refit='RMSE',
                                   return_train_score=True)
        grid_search.fit(X_train, y_train)
        cls.housing_data__grid_search = grid_search
        best_model = grid_search.best_estimator_
        predicted_values = best_model.predict(X_test)
        cls.housing_data__y_test = y_test
        cls.housing_data__y_predictions = predicted_values

    def test_MLExperimentResults_invalid_param_name_mapping(self):
        # the keys passed to parameter_name_mappings should match the parameters founds
        self.assertRaises(HelpskAssertionError,
                          lambda: MLExperimentResults.
                          from_sklearn_search_cv(searcher=self.credit_data__grid_search,
                                                 higher_score_is_better=True,
                                                 description="test description",
                                                 parameter_name_mappings={'this_should_fail': 'value'}))

    def test_MLExperimentResults_gridsearch_classification(self):
        new_param_column_names = {'model__max_features': 'max_features',
                                  'model__n_estimators': 'n_estimators',
                                  'model__min_samples_split': 'min_samples_split',
                                  'preparation__non_numeric_pipeline__encoder_chooser__transformer':
                                      'encoder'}

        # test grid search object that has multiple scores (classification)
        # passing in parameter mappings
        grid_search_credit = self.credit_data__grid_search

        parser = MLExperimentResults.from_sklearn_search_cv(searcher=grid_search_credit,
                                                            higher_score_is_better=True,
                                                            description="test description",
                                                            parameter_name_mappings=new_param_column_names)

        yaml_file = get_test_path('sklearn_eval/credit_data__grid_search.yaml')
        os.remove(yaml_file)
        parser.to_yaml_file(yaml_file)
        parser_from_dict = MLExperimentResults(parser._dict)
        parser_from_yaml = MLExperimentResults.from_yaml_file(yaml_file)

        self.assertEqual(str(parser._dict), str(parser_from_dict._dict))
        self.assertEqual(str(parser._dict), str(parser_from_yaml._dict))

        self.assertEqual(parser.description, "test description")
        self.assertEqual(parser.description, parser_from_dict.description)
        self.assertEqual(parser.description, parser_from_yaml.description)
        self.assertEqual(parser.higher_score_is_better, True)
        self.assertEqual(parser.higher_score_is_better, parser_from_dict.higher_score_is_better)
        self.assertEqual(parser.higher_score_is_better, parser_from_yaml.higher_score_is_better)
        self.assertEqual(parser.cross_validation_type,
                         "<class 'sklearn.model_selection._search.GridSearchCV'>")
        self.assertEqual(parser.cross_validation_type, parser_from_dict.cross_validation_type)
        self.assertEqual(parser.cross_validation_type, parser_from_yaml.cross_validation_type)
        self.assertEqual(parser.number_of_splits,
                         grid_search_credit.cv.n_repeats * grid_search_credit.n_splits_)
        self.assertEqual(parser.number_of_splits, parser_from_dict.number_of_splits)
        self.assertEqual(parser.number_of_splits, parser_from_yaml.number_of_splits)
        self.assertEqual(parser.score_names, ['ROC/AUC', 'F1', 'Pos. Pred. Val', 'True Pos. Rate'])
        self.assertEqual(parser.score_names, parser_from_dict.score_names)
        self.assertEqual(parser.score_names, parser_from_yaml.score_names)
        self.assertEqual(parser.parameter_names_original, list(new_param_column_names.keys()))
        self.assertEqual(parser.parameter_names_original, parser_from_dict.parameter_names_original)
        self.assertEqual(parser.parameter_names_original, parser_from_yaml.parameter_names_original)
        self.assertEqual(parser.parameter_names, list(new_param_column_names.values()))
        self.assertEqual(parser.parameter_names, parser_from_dict.parameter_names)
        self.assertEqual(parser.parameter_names, parser_from_yaml.parameter_names)
        self.assertEqual(parser.parameter_names_mapping, new_param_column_names)
        self.assertEqual(parser.parameter_names_mapping, parser_from_dict.parameter_names_mapping)
        self.assertEqual(parser.parameter_names_mapping, parser_from_yaml.parameter_names_mapping)

        self.assertTrue(isinstance(parser.test_score_rankings, dict))
        self.assertEqual(list(parser.test_score_rankings.keys()), parser.score_names)
        self.assertEqual(parser.test_score_rankings, parser_from_dict.test_score_rankings)
        self.assertEqual(parser.test_score_rankings, parser_from_yaml.test_score_rankings)
        for score in parser.score_names:
            self.assertTrue(all(np.array(parser.test_score_rankings[score]) == grid_search_credit.cv_results_[f'rank_test_{score}']))  # noqa

        self.assertEqual(
            parser.trial_labels(order_from_best_to_worst=True),
            [
                '{max_features: auto, n_estimators: 50, min_samples_split: 2, encoder: OneHotEncoder()}',
                '{max_features: auto, n_estimators: 50, min_samples_split: 2, encoder: CustomOrdinalEncoder()}',  # noqa
                '{max_features: auto, n_estimators: 10, min_samples_split: 2, encoder: OneHotEncoder()}',
                '{max_features: auto, n_estimators: 10, min_samples_split: 2, encoder: CustomOrdinalEncoder()}',  # noqa
                '{max_features: 100, n_estimators: 10, min_samples_split: 2, encoder: OneHotEncoder()}',
                '{max_features: 100, n_estimators: 10, min_samples_split: 2, encoder: CustomOrdinalEncoder()}',  # noqa
                '{max_features: 100, n_estimators: 50, min_samples_split: 2, encoder: OneHotEncoder()}',
                '{max_features: 100, n_estimators: 50, min_samples_split: 2, encoder: CustomOrdinalEncoder()}'
            ]
        )
        self.assertEqual(
            parser.trial_labels(order_from_best_to_worst=False),
            [
                '{max_features: 100, n_estimators: 10, min_samples_split: 2, encoder: OneHotEncoder()}',
                '{max_features: 100, n_estimators: 10, min_samples_split: 2, encoder: CustomOrdinalEncoder()}',  # noqa
                '{max_features: 100, n_estimators: 50, min_samples_split: 2, encoder: OneHotEncoder()}',
                '{max_features: 100, n_estimators: 50, min_samples_split: 2, encoder: CustomOrdinalEncoder()}',  # noqa
                '{max_features: auto, n_estimators: 10, min_samples_split: 2, encoder: OneHotEncoder()}',
                '{max_features: auto, n_estimators: 10, min_samples_split: 2, encoder: CustomOrdinalEncoder()}',  # noqa
                '{max_features: auto, n_estimators: 50, min_samples_split: 2, encoder: OneHotEncoder()}',
                '{max_features: auto, n_estimators: 50, min_samples_split: 2, encoder: CustomOrdinalEncoder()}'  # noqa
             ]
        )

        def assert_np_arrays_are_close(array1, array2):
            self.assertEqual(len(array1), len(array2))
            for index in range(len(array1)):
                is_close = hlp.validation.is_close(array1[index], array2[index])
                both_nan = np.isnan(array1[index]) and np.isnan(array2[index])
                self.assertTrue(is_close or both_nan)

        self.assertTrue(isinstance(parser.test_score_averages, dict))
        self.assertEqual(list(parser.test_score_averages.keys()), parser.score_names)

        assert_np_arrays_are_close(np.array([1, 2, 3]), np.array([1, 2, 3]))
        assert_np_arrays_are_close(np.array([1, 2, np.nan]), np.array([1, 2, np.nan]))
        self.assertRaises(AssertionError,
                          lambda: assert_np_arrays_are_close(np.array([1, 2, 3]), np.array([1, 2, 3.001])))
        self.assertRaises(AssertionError,
                          lambda: assert_np_arrays_are_close(np.array([1, 2, 3]), np.array([1, 2, np.nan])))

        cv_dataframe = parser.to_dataframe(exclude_zero_variance_params=False)
        with redirect_stdout_to_file(get_test_path('sklearn_eval/credit__grid_search__all_scores__dataframe__no_exclude.txt')):  # noqa
            print_dataframe(cv_dataframe)
        # ensure the hyper-param columns (last 4 columns) are in the same order as the mapping.
        self.assertEqual(list(cv_dataframe.columns[-4:]), list(new_param_column_names.values()))

        self.assertTrue('min_samples_split' in cv_dataframe.columns)
        self.assertTrue(all(cv_dataframe['min_samples_split'] == [2, 2, 2, 2, 2, 2, 2, 2]))
        del cv_dataframe

        with redirect_stdout_to_file(get_test_path('sklearn_eval/credit__grid_search__all_scores__dataframe__exclude.txt')):  # noqa
            print_dataframe(parser.to_dataframe(exclude_zero_variance_params=True))

        self.assertEqual(list(parser.best_trial_indexes), list(parser.to_dataframe().index))
        cv_dataframe = parser.to_dataframe().sort_index()
        # ensure the hyper-param columns (now last 3 columns) are in the same order as the mapping
        self.assertEqual(list(cv_dataframe.columns[-3:]), ['max_features', 'n_estimators', 'encoder'])
        self.assertFalse('min_samples_split' in cv_dataframe.columns)
        hlp.validation.assert_dataframes_match([parser.to_dataframe(sort_by_score=False), cv_dataframe])

        labeled_dataframe = parser.to_labeled_dataframe()
        self.assertTrue(all(labeled_dataframe['Trial Index'] == list(range(1, parser.number_of_trials + 1))))
        self.assertIsNotNone(labeled_dataframe['label'])
        hlp.validation.assert_dataframes_match(dataframes=[parser.to_dataframe(sort_by_score=False),
                                                           labeled_dataframe.drop(columns=['Trial Index',
                                                                                           'label'])])

        assert_np_arrays_are_close(cv_dataframe[f'{parser.score_names[0]} Mean'],
                                   grid_search_credit.cv_results_[f'mean_test_{parser.score_names[0]}'])
        assert_np_arrays_are_close(cv_dataframe[f'{parser.score_names[1]} Mean'],
                                   grid_search_credit.cv_results_[f'mean_test_{parser.score_names[1]}'])
        assert_np_arrays_are_close(cv_dataframe[f'{parser.score_names[2]} Mean'],
                                   grid_search_credit.cv_results_[f'mean_test_{parser.score_names[2]}'])
        assert_np_arrays_are_close(cv_dataframe[f'{parser.score_names[3]} Mean'],
                                   grid_search_credit.cv_results_[f'mean_test_{parser.score_names[3]}'])

        self.assertEqual(list(grid_search_credit.cv_results_['param_model__max_features'].data),
                         cv_dataframe[parser.parameter_names[0]].tolist())
        self.assertEqual(list(grid_search_credit.cv_results_['param_model__n_estimators'].data),
                         cv_dataframe[parser.parameter_names[1]].tolist())
        encoder_param_name = 'param_preparation__non_numeric_pipeline__encoder_chooser__transformer'
        encoder_values = [str(x) for x in list(grid_search_credit.cv_results_[encoder_param_name].data)]
        self.assertEqual(encoder_values, cv_dataframe[parser.parameter_names[3]].tolist())

        with open(get_test_path('sklearn_eval/credit__grid_search__all_scores.html'), 'w') as file:
            file.write(clean_formatted_dataframe(parser.to_formatted_dataframe().to_html()))

        with open(get_test_path('sklearn_eval/credit__grid_search__all_scores_2.html'), 'w') as file:
            file.write(clean_formatted_dataframe(parser.to_formatted_dataframe(exclude_zero_variance_params=False).to_html()))  # noqa

        with open(get_test_path('sklearn_eval/credit__grid_search__primary_score_only.html'), 'w') as file:
            file.write(clean_formatted_dataframe(parser.to_formatted_dataframe(primary_score_only=True).to_html()))  # noqa

        assert_np_arrays_are_close(parser.primary_score_averages,
                                   np.array(parser.test_score_averages[parser.primary_score_name]))
        assert_np_arrays_are_close(parser.primary_score_averages,
                                   grid_search_credit.cv_results_['mean_test_ROC/AUC'])
        assert_np_arrays_are_close(parser.primary_score_averages,
                                   parser_from_dict.primary_score_averages)
        assert_np_arrays_are_close(parser.primary_score_averages,
                                   parser_from_yaml.primary_score_averages)

        self.assertEqual(len(parser.score_standard_errors(score_name=parser.score_names[0])),
                         parser.number_of_trials)
        self.assertEqual(len(parser.score_standard_errors(score_name=parser.score_names[1])),
                         parser.number_of_trials)
        self.assertEqual(len(parser.score_standard_errors(score_name=parser.score_names[2])),
                         parser.number_of_trials)
        self.assertEqual(len(parser.score_standard_errors(score_name=parser.score_names[3])),
                         parser.number_of_trials)

        self.assertEqual(parser.numeric_parameters, ['n_estimators'])
        self.assertEqual(parser.non_numeric_parameters, ['max_features', 'encoder'])

        for score in parser.score_names:
            assert_np_arrays_are_close(np.array(parser.test_score_averages[score]),
                                       grid_search_credit.cv_results_[f'mean_test_{score}'])
            assert_np_arrays_are_close(np.array(parser.test_score_averages[score]),
                                       np.array(parser_from_dict.test_score_averages[score]))
            assert_np_arrays_are_close(np.array(parser.test_score_averages[score]),
                                       np.array(parser_from_yaml.test_score_averages[score]))

        self.assertTrue(isinstance(parser.test_score_standard_deviations, dict))
        self.assertEqual(list(parser.test_score_standard_deviations.keys()), parser.score_names)

        for score in parser.score_names:
            assert_np_arrays_are_close(np.array(parser.test_score_standard_deviations[score]),
                                       grid_search_credit.cv_results_[f'std_test_{score}'])
            assert_np_arrays_are_close(np.array(parser.test_score_standard_deviations[score]),
                                       np.array(parser_from_dict.test_score_standard_deviations[score]))
            assert_np_arrays_are_close(np.array(parser.test_score_standard_deviations[score]),
                                       np.array(parser_from_yaml.test_score_standard_deviations[score]))

        self.assertTrue(isinstance(parser.train_score_averages, dict))
        self.assertEqual(list(parser.train_score_averages.keys()), parser.score_names)

        for score in parser.score_names:
            assert_np_arrays_are_close(np.array(parser.train_score_averages[score]),
                                       grid_search_credit.cv_results_[f'mean_train_{score}'])
            assert_np_arrays_are_close(np.array(parser.train_score_averages[score]),
                                       np.array(parser_from_dict.train_score_averages[score]))
            assert_np_arrays_are_close(np.array(parser.train_score_averages[score]),
                                       np.array(parser_from_yaml.train_score_averages[score]))

        self.assertTrue(isinstance(parser.train_score_standard_deviations, dict))
        self.assertEqual(list(parser.train_score_standard_deviations.keys()), parser.score_names)

        for score in parser.score_names:
            assert_np_arrays_are_close(np.array(parser.train_score_standard_deviations[score]),
                                       grid_search_credit.cv_results_[f'std_train_{score}'])
            assert_np_arrays_are_close(np.array(parser.train_score_standard_deviations[score]),
                                       np.array(parser_from_dict.train_score_standard_deviations[score]))
            assert_np_arrays_are_close(np.array(parser.train_score_standard_deviations[score]),
                                       np.array(parser_from_yaml.train_score_standard_deviations[score]))

        self.assertTrue(isinstance(parser.trials, list))
        self.assertEqual(len(parser.trials), parser.number_of_trials)
        self.assertEqual(parser.trials, parser_from_dict.trials)
        self.assertEqual(parser.trials, parser_from_yaml.trials)

        self.assertTrue(isinstance(parser.timings, dict))
        self.assertEqual(list(parser.timings.keys()), ['fit time averages',
                                                       'fit time standard deviations',
                                                       'score time averages',
                                                       'score time standard deviations'])

        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[0]]),
                                   grid_search_credit.cv_results_['mean_fit_time'])
        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[1]]),
                                   grid_search_credit.cv_results_['std_fit_time'])
        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[2]]),
                                   grid_search_credit.cv_results_['mean_score_time'])
        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[3]]),
                                   grid_search_credit.cv_results_['std_score_time'])

        for timing in list(parser.timings.keys()):
            assert_np_arrays_are_close(np.array(parser.timings[timing]),
                                       np.array(parser_from_dict.timings[timing]))
            assert_np_arrays_are_close(np.array(parser.timings[timing]),
                                       np.array(parser_from_yaml.timings[timing]))

        self.assertEqual(parser.number_of_trials, 8)
        self.assertEqual(parser.number_of_trials, parser_from_dict.number_of_trials)
        self.assertEqual(parser.number_of_trials, parser_from_yaml.number_of_trials)

        self.assertEqual(parser.number_of_scores, 4)
        self.assertEqual(parser.number_of_scores, parser_from_dict.number_of_scores)
        self.assertEqual(parser.number_of_scores, parser_from_yaml.number_of_scores)

        self.assertEqual(parser.primary_score_name, 'ROC/AUC')
        self.assertEqual(parser.primary_score_name, parser_from_dict.primary_score_name)
        self.assertEqual(parser.primary_score_name, parser_from_yaml.primary_score_name)

        self.assertTrue(all(parser.trial_rankings == grid_search_credit.cv_results_['rank_test_ROC/AUC']))  # noqa
        self.assertEqual(parser.best_score_index, 6)
        self.assertEqual(parser.best_score,
                         grid_search_credit.cv_results_['mean_test_ROC/AUC'][6])
        self.assertEqual(parser.best_score,
                         np.nanmax(grid_search_credit.cv_results_['mean_test_ROC/AUC']))

        self.assertEqual(parser.parameter_names_original,
                         ['model__max_features',
                          'model__n_estimators',
                          'model__min_samples_split',
                          'preparation__non_numeric_pipeline__encoder_chooser__transformer'])
        self.assertEqual(parser.parameter_names,
                         ['max_features', 'n_estimators', 'min_samples_split', 'encoder'])
        self.assertEqual(parser.best_params,
                         {'max_features': 'auto',
                          'n_estimators': 50,
                          'min_samples_split': 2,
                          'encoder': 'OneHotEncoder()'})

        self.assertTrue(all([parser.to_dataframe(exclude_zero_variance_params=False).loc[parser.best_score_index, key] == value  # noqa
                             for key, value in parser.best_params.items()]))

        self.assertEqual(parser.best_score, parser_from_dict.best_score)
        self.assertEqual(parser.best_score, parser_from_yaml.best_score)

    def test_MLExperimentResults_gridsearch_classification_single_score(self):
        # test grid search object that has one score (classification)
        # not passing in parameter mappings
        grid_search_credit = self.credit_data__grid_search__roc_auc
        parser = MLExperimentResults.from_sklearn_search_cv(searcher=grid_search_credit,
                                                            higher_score_is_better=True,
                                                            description="test description",
                                                            parameter_name_mappings=None)
        yaml_file = get_test_path('sklearn_eval/credit_data__grid_search_roc.yaml')
        os.remove(yaml_file)
        parser.to_yaml_file(yaml_file)
        parser_from_dict = MLExperimentResults(parser._dict)
        parser_from_yaml = MLExperimentResults.from_yaml_file(yaml_file)

        self.assertEqual(str(parser._dict), str(parser_from_dict._dict))
        self.assertEqual(str(parser._dict), str(parser_from_yaml._dict))

        self.assertEqual(parser.description, "test description")
        self.assertEqual(parser.description, parser_from_dict.description)
        self.assertEqual(parser.description, parser_from_yaml.description)
        self.assertEqual(parser.higher_score_is_better, True)
        self.assertEqual(parser.higher_score_is_better, parser_from_dict.higher_score_is_better)
        self.assertEqual(parser.higher_score_is_better, parser_from_yaml.higher_score_is_better)
        self.assertEqual(parser.cross_validation_type,
                         "<class 'sklearn.model_selection._search.GridSearchCV'>")
        self.assertEqual(parser.cross_validation_type, parser_from_dict.cross_validation_type)
        self.assertEqual(parser.cross_validation_type, parser_from_yaml.cross_validation_type)
        self.assertEqual(parser.number_of_splits,
                         grid_search_credit.cv.n_repeats * grid_search_credit.n_splits_)
        self.assertEqual(parser.number_of_splits, parser_from_dict.number_of_splits)
        self.assertEqual(parser.number_of_splits, parser_from_yaml.number_of_splits)
        self.assertEqual(parser.score_names, ['roc_auc'])
        self.assertEqual(parser.score_names, parser_from_dict.score_names)
        self.assertEqual(parser.score_names, parser_from_yaml.score_names)
        self.assertEqual(parser.parameter_names_original,
                         ['model__max_features', 'model__n_estimators',
                          'preparation__non_numeric_pipeline__encoder_chooser__transformer'])
        self.assertEqual(parser.parameter_names_original, parser_from_dict.parameter_names_original)
        self.assertEqual(parser.parameter_names_original, parser_from_yaml.parameter_names_original)
        self.assertEqual(parser.parameter_names, parser.parameter_names_original)
        self.assertEqual(parser.parameter_names, parser_from_dict.parameter_names)
        self.assertEqual(parser.parameter_names, parser_from_yaml.parameter_names)
        self.assertIsNone(parser.parameter_names_mapping)
        self.assertIsNone(parser_from_dict.parameter_names_mapping)
        self.assertIsNone(parser_from_yaml.parameter_names_mapping)

        self.assertTrue(isinstance(parser.test_score_rankings, dict))
        self.assertEqual(list(parser.test_score_rankings.keys()), parser.score_names)
        self.assertTrue(all(np.array(parser.test_score_rankings['roc_auc']) == grid_search_credit.cv_results_['rank_test_score']))  # noqa
        self.assertEqual(parser.test_score_rankings, parser_from_dict.test_score_rankings)
        self.assertEqual(parser.test_score_rankings, parser_from_yaml.test_score_rankings)

        self.assertEqual(parser.trial_labels(order_from_best_to_worst=True),
                         ['{model__max_features: auto, model__n_estimators: 50, preparation__non_numeric_pipeline__encoder_chooser__transformer: OneHotEncoder()}',  # noqa
                          '{model__max_features: auto, model__n_estimators: 50, preparation__non_numeric_pipeline__encoder_chooser__transformer: CustomOrdinalEncoder()}',  # noqa
                          '{model__max_features: auto, model__n_estimators: 10, preparation__non_numeric_pipeline__encoder_chooser__transformer: OneHotEncoder()}',  # noqa
                          '{model__max_features: auto, model__n_estimators: 10, preparation__non_numeric_pipeline__encoder_chooser__transformer: CustomOrdinalEncoder()}',  # noqa
                          '{model__max_features: 100, model__n_estimators: 10, preparation__non_numeric_pipeline__encoder_chooser__transformer: OneHotEncoder()}',  # noqa
                          '{model__max_features: 100, model__n_estimators: 10, preparation__non_numeric_pipeline__encoder_chooser__transformer: CustomOrdinalEncoder()}',  # noqa
                          '{model__max_features: 100, model__n_estimators: 50, preparation__non_numeric_pipeline__encoder_chooser__transformer: OneHotEncoder()}',  # noqa
                          '{model__max_features: 100, model__n_estimators: 50, preparation__non_numeric_pipeline__encoder_chooser__transformer: CustomOrdinalEncoder()}'])  # noqa
        self.assertEqual(parser.trial_labels(order_from_best_to_worst=False),
                         ['{model__max_features: 100, model__n_estimators: 10, preparation__non_numeric_pipeline__encoder_chooser__transformer: OneHotEncoder()}',  # noqa
                          '{model__max_features: 100, model__n_estimators: 10, preparation__non_numeric_pipeline__encoder_chooser__transformer: CustomOrdinalEncoder()}',  # noqa
                          '{model__max_features: 100, model__n_estimators: 50, preparation__non_numeric_pipeline__encoder_chooser__transformer: OneHotEncoder()}',  # noqa
                          '{model__max_features: 100, model__n_estimators: 50, preparation__non_numeric_pipeline__encoder_chooser__transformer: CustomOrdinalEncoder()}',  # noqa
                          '{model__max_features: auto, model__n_estimators: 10, preparation__non_numeric_pipeline__encoder_chooser__transformer: OneHotEncoder()}',  # noqa
                          '{model__max_features: auto, model__n_estimators: 10, preparation__non_numeric_pipeline__encoder_chooser__transformer: CustomOrdinalEncoder()}',  # noqa
                          '{model__max_features: auto, model__n_estimators: 50, preparation__non_numeric_pipeline__encoder_chooser__transformer: OneHotEncoder()}',  # noqa
                          '{model__max_features: auto, model__n_estimators: 50, preparation__non_numeric_pipeline__encoder_chooser__transformer: CustomOrdinalEncoder()}'])  # noqa

        def assert_np_arrays_are_close(array1, array2):
            self.assertEqual(len(array1), len(array2))
            for index in range(len(array1)):
                is_close = hlp.validation.is_close(array1[index], array2[index])
                both_nan = np.isnan(array1[index]) and np.isnan(array2[index])
                self.assertTrue(is_close or both_nan)

        assert_np_arrays_are_close(np.array(parser.test_score_averages[parser.primary_score_name]),
                                   grid_search_credit.cv_results_['mean_test_score'])
        assert_np_arrays_are_close(np.array(parser.test_score_averages[parser.primary_score_name]),
                                   np.array(parser_from_dict.test_score_averages[parser.primary_score_name]))
        assert_np_arrays_are_close(np.array(parser.test_score_averages[parser.primary_score_name]),
                                   np.array(parser_from_yaml.test_score_averages[parser.primary_score_name]))

        self.assertEqual(list(parser.best_trial_indexes), list(parser.to_dataframe().index))
        with redirect_stdout_to_file(get_test_path('sklearn_eval/credit__grid_search__single_score__dataframe.txt')):  # noqa
            print_dataframe(parser.to_dataframe())
        cv_dataframe = parser.to_dataframe().sort_index()
        hlp.validation.assert_dataframes_match([parser.to_dataframe(sort_by_score=False), cv_dataframe])
        assert_np_arrays_are_close(cv_dataframe[f'{parser.score_names[0]} Mean'],
                                   grid_search_credit.cv_results_['mean_test_score'])

        labeled_dataframe = parser.to_labeled_dataframe()
        self.assertTrue(all(labeled_dataframe['Trial Index'] == list(range(1, parser.number_of_trials + 1))))
        self.assertIsNotNone(labeled_dataframe['label'])
        hlp.validation.assert_dataframes_match(dataframes=[parser.to_dataframe(sort_by_score=False),
                                                           labeled_dataframe.drop(columns=['Trial Index',
                                                                                           'label'])])

        self.assertEqual(list(grid_search_credit.cv_results_['param_model__max_features'].data), cv_dataframe[parser.parameter_names[0]].tolist())  # noqa
        self.assertEqual(list(grid_search_credit.cv_results_['param_model__n_estimators'].data), cv_dataframe[parser.parameter_names[1]].tolist())  # noqa
        encoder_values = [str(x) for x in
                          list(grid_search_credit.cv_results_['param_preparation__non_numeric_pipeline__encoder_chooser__transformer'].data)]  # noqa
        self.assertEqual(encoder_values, cv_dataframe[parser.parameter_names[2]].tolist())

        with open(get_test_path('sklearn_eval/credit__grid_search__single_score_all_scores.html'), 'w') as file:  # noqa
            file.write(clean_formatted_dataframe(parser.to_formatted_dataframe().to_html()))

        with open(get_test_path('sklearn_eval/credit__grid_search__single_score_primary_score_only.html'), 'w') as file:  # noqa
            file.write(clean_formatted_dataframe(parser.to_formatted_dataframe(primary_score_only=True).to_html()))  # noqa

        self.assertTrue(isinstance(parser.test_score_averages, dict))
        self.assertEqual(list(parser.test_score_averages.keys()), parser.score_names)

        assert_np_arrays_are_close(np.array([1, 2, 3]), np.array([1, 2, 3]))
        assert_np_arrays_are_close(np.array([1, 2, np.nan]), np.array([1, 2, np.nan]))
        self.assertRaises(AssertionError,
                          lambda: assert_np_arrays_are_close(np.array([1, 2, 3]), np.array([1, 2, 3.001])))
        self.assertRaises(AssertionError,
                          lambda: assert_np_arrays_are_close(np.array([1, 2, 3]), np.array([1, 2, np.nan])))

        assert_np_arrays_are_close(parser.primary_score_averages,
                                   np.array(parser.test_score_averages[parser.primary_score_name]))
        assert_np_arrays_are_close(parser.primary_score_averages,
                                   grid_search_credit.cv_results_['mean_test_score'])
        assert_np_arrays_are_close(parser.primary_score_averages,
                                   parser_from_dict.primary_score_averages)
        assert_np_arrays_are_close(parser.primary_score_averages,
                                   parser_from_yaml.primary_score_averages)

        self.assertEqual(len(parser.score_standard_errors(score_name=parser.score_names[0])),
                         parser.number_of_trials)

        self.assertEqual(parser.numeric_parameters, ['model__n_estimators'])
        self.assertEqual(
            parser.non_numeric_parameters,
            ['model__max_features', 'preparation__non_numeric_pipeline__encoder_chooser__transformer']
        )

        assert_np_arrays_are_close(np.array(parser.test_score_averages[parser.primary_score_name]),
                                   grid_search_credit.cv_results_['mean_test_score'])
        assert_np_arrays_are_close(np.array(parser.test_score_averages[parser.primary_score_name]),
                                   np.array(parser_from_dict.test_score_averages[parser.primary_score_name]))
        assert_np_arrays_are_close(np.array(parser.test_score_averages[parser.primary_score_name]),
                                   np.array(parser_from_yaml.test_score_averages[parser.primary_score_name]))

        self.assertTrue(isinstance(parser.test_score_standard_deviations, dict))
        self.assertEqual(list(parser.test_score_standard_deviations.keys()), parser.score_names)

        assert_np_arrays_are_close(np.array(parser.test_score_standard_deviations[parser.primary_score_name]),
                                   grid_search_credit.cv_results_['std_test_score'])
        assert_np_arrays_are_close(np.array(parser.test_score_standard_deviations[parser.primary_score_name]),
                                   np.array(parser_from_dict.test_score_standard_deviations[parser.primary_score_name]))  # noqa
        assert_np_arrays_are_close(np.array(parser.test_score_standard_deviations[parser.primary_score_name]),
                                   np.array(parser_from_yaml.test_score_standard_deviations[parser.primary_score_name]))  # noqa

        self.assertIsNone(parser.train_score_averages)
        self.assertIsNone(parser.train_score_standard_deviations)

        self.assertTrue(isinstance(parser.trials, list))
        self.assertEqual(len(parser.trials), parser.number_of_trials)
        self.assertEqual(parser.trials, parser_from_dict.trials)
        self.assertEqual(parser.trials, parser_from_yaml.trials)

        self.assertTrue(isinstance(parser.timings, dict))
        self.assertEqual(list(parser.timings.keys()), ['fit time averages',
                                                       'fit time standard deviations',
                                                       'score time averages',
                                                       'score time standard deviations'])

        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[0]]),
                                   grid_search_credit.cv_results_['mean_fit_time'])
        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[1]]),
                                   grid_search_credit.cv_results_['std_fit_time'])
        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[2]]),
                                   grid_search_credit.cv_results_['mean_score_time'])
        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[3]]),
                                   grid_search_credit.cv_results_['std_score_time'])

        for timing in list(parser.timings.keys()):
            assert_np_arrays_are_close(np.array(parser.timings[timing]),
                                       np.array(parser_from_dict.timings[timing]))
            assert_np_arrays_are_close(np.array(parser.timings[timing]),
                                       np.array(parser_from_yaml.timings[timing]))

        self.assertEqual(parser.number_of_trials, 8)
        self.assertEqual(parser.number_of_trials, parser_from_dict.number_of_trials)
        self.assertEqual(parser.number_of_trials, parser_from_yaml.number_of_trials)

        self.assertEqual(parser.number_of_scores, 1)
        self.assertEqual(parser.number_of_scores, parser_from_dict.number_of_scores)
        self.assertEqual(parser.number_of_scores, parser_from_yaml.number_of_scores)

        self.assertEqual(parser.primary_score_name, 'roc_auc')
        self.assertEqual(parser.primary_score_name, parser_from_dict.primary_score_name)
        self.assertEqual(parser.primary_score_name, parser_from_yaml.primary_score_name)

        self.assertTrue(all(parser.trial_rankings == grid_search_credit.cv_results_['rank_test_score']))  # noqa
        self.assertEqual(parser.best_score_index, 6)
        self.assertEqual(parser.best_score,
                         grid_search_credit.cv_results_['mean_test_score'][6])
        self.assertEqual(parser.best_score,
                         np.nanmax(grid_search_credit.cv_results_['mean_test_score']))

        self.assertTrue(all([parser.to_dataframe().loc[parser.best_score_index, key] == value
                             for key, value in parser.best_params.items()]))

        self.assertEqual(parser.best_params['model__max_features'],
                         grid_search_credit.best_params_['model__max_features'])
        self.assertEqual(parser.best_params['model__n_estimators'],
                         grid_search_credit.best_params_['model__n_estimators'])
        self.assertEqual(
            parser.best_params['preparation__non_numeric_pipeline__encoder_chooser__transformer'],
            str(grid_search_credit.best_params_['preparation__non_numeric_pipeline__encoder_chooser__transformer'])  # noqa
        )

        self.assertEqual(parser.best_score, parser_from_dict.best_score)
        self.assertEqual(parser.best_score, parser_from_yaml.best_score)

    def test_MLExperimentResults_gridsearch_regression(self):
        # test grid search object that has multiple scores (regression)
        # not passing in parameter mappings
        grid_search_housing = self.housing_data__grid_search
        parser = MLExperimentResults.from_sklearn_search_cv(searcher=grid_search_housing,
                                                            higher_score_is_better=False,
                                                            description="test description")
        yaml_file = get_test_path('sklearn_eval/housing_data__grid_search.yaml')
        os.remove(yaml_file)
        parser.to_yaml_file(yaml_file)
        parser_from_dict = MLExperimentResults(parser._dict)
        parser_from_yaml = MLExperimentResults.from_yaml_file(yaml_file)

        self.assertEqual(str(parser._dict), str(parser_from_dict._dict))
        self.assertEqual(str(parser._dict), str(parser_from_yaml._dict))

        self.assertEqual(parser.description, "test description")
        self.assertEqual(parser.description, parser_from_dict.description)
        self.assertEqual(parser.description, parser_from_yaml.description)
        self.assertEqual(parser.higher_score_is_better, False)
        self.assertEqual(parser.higher_score_is_better, parser_from_dict.higher_score_is_better)
        self.assertEqual(parser.higher_score_is_better, parser_from_yaml.higher_score_is_better)
        self.assertEqual(parser.cross_validation_type,
                         "<class 'sklearn.model_selection._search.GridSearchCV'>")
        self.assertEqual(parser.cross_validation_type, parser_from_dict.cross_validation_type)
        self.assertEqual(parser.cross_validation_type, parser_from_yaml.cross_validation_type)
        self.assertEqual(parser.number_of_splits,
                         grid_search_housing.cv.n_repeats * grid_search_housing.n_splits_)
        self.assertEqual(parser.number_of_splits, parser_from_dict.number_of_splits)
        self.assertEqual(parser.number_of_splits, parser_from_yaml.number_of_splits)
        self.assertEqual(parser.score_names, ['RMSE', 'MAE'])
        self.assertEqual(parser.score_names, parser_from_dict.score_names)
        self.assertEqual(parser.score_names, parser_from_yaml.score_names)

        self.assertEqual(parser.parameter_names_original, parser_from_dict.parameter_names_original)
        self.assertEqual(parser.parameter_names_original, parser_from_yaml.parameter_names_original)
        self.assertEqual(parser.parameter_names, parser.parameter_names_original)
        self.assertEqual(parser.parameter_names, parser_from_dict.parameter_names)
        self.assertEqual(parser.parameter_names, parser_from_yaml.parameter_names)
        self.assertIsNone(parser.parameter_names_mapping)
        self.assertIsNone(parser_from_dict.parameter_names_mapping)
        self.assertIsNone(parser_from_yaml.parameter_names_mapping)

        self.assertTrue(isinstance(parser.test_score_rankings, dict))
        self.assertEqual(list(parser.test_score_rankings.keys()), parser.score_names)
        self.assertEqual(parser.test_score_rankings, parser_from_dict.test_score_rankings)
        self.assertEqual(parser.test_score_rankings, parser_from_yaml.test_score_rankings)
        for score in parser.score_names:
            self.assertTrue(all(np.array(parser.test_score_rankings[score]) == grid_search_housing.cv_results_[f'rank_test_{score}']))  # noqa

        self.assertEqual(parser.trial_labels(order_from_best_to_worst=True),
                         ['{model__max_features: auto, model__n_estimators: 50}',
                          '{model__max_features: auto, model__n_estimators: 10}',
                          '{model__max_features: 2, model__n_estimators: 50}',
                          '{model__max_features: 2, model__n_estimators: 10}'])
        self.assertEqual(parser.trial_labels(order_from_best_to_worst=False),
                         ['{model__max_features: 2, model__n_estimators: 10}',
                          '{model__max_features: 2, model__n_estimators: 50}',
                          '{model__max_features: auto, model__n_estimators: 10}',
                          '{model__max_features: auto, model__n_estimators: 50}'])

        def assert_np_arrays_are_close(array1, array2):
            self.assertEqual(len(array1), len(array2))
            for index in range(len(array1)):
                is_close = hlp.validation.is_close(array1[index], array2[index])
                both_nan = np.isnan(array1[index]) and np.isnan(array2[index])
                self.assertTrue(is_close or both_nan)

        self.assertTrue(isinstance(parser.test_score_averages, dict))
        self.assertEqual(list(parser.test_score_averages.keys()), parser.score_names)

        assert_np_arrays_are_close(np.array([1, 2, 3]), np.array([1, 2, 3]))
        assert_np_arrays_are_close(np.array([1, 2, np.nan]), np.array([1, 2, np.nan]))
        self.assertRaises(AssertionError,
                          lambda: assert_np_arrays_are_close(np.array([1, 2, 3]), np.array([1, 2, 3.001])))
        self.assertRaises(AssertionError,
                          lambda: assert_np_arrays_are_close(np.array([1, 2, 3]), np.array([1, 2, np.nan])))

        self.assertEqual(list(parser.best_trial_indexes), list(parser.to_dataframe().index))
        with redirect_stdout_to_file(get_test_path('sklearn_eval/housing__grid_search__dataframe.txt')):
            print_dataframe(parser.to_dataframe())
        cv_dataframe = parser.to_dataframe().sort_index()
        hlp.validation.assert_dataframes_match([parser.to_dataframe(sort_by_score=False), cv_dataframe])
        assert_np_arrays_are_close(cv_dataframe[f'{parser.score_names[0]} Mean'],
                                   grid_search_housing.cv_results_[f'mean_test_{parser.score_names[0]}'] * -1)
        assert_np_arrays_are_close(cv_dataframe[f'{parser.score_names[1]} Mean'],
                                   grid_search_housing.cv_results_[f'mean_test_{parser.score_names[1]}'] * -1)

        labeled_dataframe = parser.to_labeled_dataframe()
        self.assertTrue(all(labeled_dataframe['Trial Index'] == list(range(1, parser.number_of_trials + 1))))
        self.assertIsNotNone(labeled_dataframe['label'])
        hlp.validation.assert_dataframes_match(dataframes=[parser.to_dataframe(sort_by_score=False),
                                                           labeled_dataframe.drop(columns=['Trial Index',
                                                                                           'label'])])

        self.assertEqual(list(grid_search_housing.cv_results_['param_model__max_features'].data), cv_dataframe[parser.parameter_names[0]].tolist())  # noqa
        self.assertEqual(list(grid_search_housing.cv_results_['param_model__n_estimators'].data), cv_dataframe[parser.parameter_names[1]].tolist())  # noqa

        with open(get_test_path('sklearn_eval/housing__grid_search__all_scores.html'), 'w') as file:
            file.write(clean_formatted_dataframe(parser.to_formatted_dataframe().to_html()))

        with open(get_test_path('sklearn_eval/housing__grid_search__primary_score_only.html'), 'w') as file:
            file.write(clean_formatted_dataframe(parser.to_formatted_dataframe(primary_score_only=True).to_html()))  # noqa

        self.assertTrue(isinstance(parser.test_score_averages, dict))
        self.assertEqual(list(parser.test_score_averages.keys()), parser.score_names)

        assert_np_arrays_are_close(parser.primary_score_averages,
                                   np.array(parser.test_score_averages[parser.primary_score_name]))
        assert_np_arrays_are_close(parser.primary_score_averages,
                                   grid_search_housing.cv_results_['mean_test_RMSE'] * -1)
        assert_np_arrays_are_close(parser.primary_score_averages,
                                   parser_from_dict.primary_score_averages)
        assert_np_arrays_are_close(parser.primary_score_averages,
                                   parser_from_yaml.primary_score_averages)

        self.assertEqual(len(parser.score_standard_errors(score_name=parser.score_names[0])),
                         parser.number_of_trials)
        self.assertEqual(len(parser.score_standard_errors(score_name=parser.score_names[1])),
                         parser.number_of_trials)

        self.assertEqual(parser.numeric_parameters, ['model__n_estimators'])
        self.assertEqual(parser.non_numeric_parameters, ['model__max_features'])

        for score in parser.score_names:
            assert_np_arrays_are_close(np.array(parser.test_score_averages[score]),
                                       grid_search_housing.cv_results_[f'mean_test_{score}'] * -1)
            assert_np_arrays_are_close(np.array(parser.test_score_averages[score]),
                                       np.array(parser_from_dict.test_score_averages[score]))
            assert_np_arrays_are_close(np.array(parser.test_score_averages[score]),
                                       np.array(parser_from_yaml.test_score_averages[score]))

        self.assertTrue(isinstance(parser.test_score_standard_deviations, dict))
        self.assertEqual(list(parser.test_score_standard_deviations.keys()), parser.score_names)

        for score in parser.score_names:
            assert_np_arrays_are_close(np.array(parser.test_score_standard_deviations[score]),
                                       grid_search_housing.cv_results_[f'std_test_{score}'])
            assert_np_arrays_are_close(np.array(parser.test_score_standard_deviations[score]),
                                       np.array(parser_from_dict.test_score_standard_deviations[score]))
            assert_np_arrays_are_close(np.array(parser.test_score_standard_deviations[score]),
                                       np.array(parser_from_yaml.test_score_standard_deviations[score]))

        self.assertTrue(isinstance(parser.train_score_averages, dict))
        self.assertEqual(list(parser.train_score_averages.keys()), parser.score_names)

        for score in parser.score_names:
            assert_np_arrays_are_close(np.array(parser.train_score_averages[score]),
                                       grid_search_housing.cv_results_[f'mean_train_{score}'] * -1)
            assert_np_arrays_are_close(np.array(parser.train_score_averages[score]),
                                       np.array(parser_from_dict.train_score_averages[score]))
            assert_np_arrays_are_close(np.array(parser.train_score_averages[score]),
                                       np.array(parser_from_yaml.train_score_averages[score]))

        self.assertTrue(isinstance(parser.train_score_standard_deviations, dict))
        self.assertEqual(list(parser.train_score_standard_deviations.keys()), parser.score_names)

        for score in parser.score_names:
            assert_np_arrays_are_close(np.array(parser.train_score_standard_deviations[score]),
                                       grid_search_housing.cv_results_[f'std_train_{score}'])
            assert_np_arrays_are_close(np.array(parser.train_score_standard_deviations[score]),
                                       np.array(parser_from_dict.train_score_standard_deviations[score]))
            assert_np_arrays_are_close(np.array(parser.train_score_standard_deviations[score]),
                                       np.array(parser_from_yaml.train_score_standard_deviations[score]))

        self.assertTrue(isinstance(parser.trials, list))
        self.assertEqual(len(parser.trials), parser.number_of_trials)
        self.assertEqual(parser.trials, parser_from_dict.trials)
        self.assertEqual(parser.trials, parser_from_yaml.trials)

        self.assertTrue(isinstance(parser.timings, dict))
        self.assertEqual(list(parser.timings.keys()), ['fit time averages',
                                                       'fit time standard deviations',
                                                       'score time averages',
                                                       'score time standard deviations'])

        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[0]]),
                                   grid_search_housing.cv_results_['mean_fit_time'])
        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[1]]),
                                   grid_search_housing.cv_results_['std_fit_time'])
        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[2]]),
                                   grid_search_housing.cv_results_['mean_score_time'])
        assert_np_arrays_are_close(np.array(parser.timings[list(parser.timings.keys())[3]]),
                                   grid_search_housing.cv_results_['std_score_time'])

        for timing in list(parser.timings.keys()):
            assert_np_arrays_are_close(np.array(parser.timings[timing]),
                                       np.array(parser_from_dict.timings[timing]))
            assert_np_arrays_are_close(np.array(parser.timings[timing]),
                                       np.array(parser_from_yaml.timings[timing]))

        self.assertEqual(parser.number_of_trials, 4)
        self.assertEqual(parser.number_of_trials, parser_from_dict.number_of_trials)
        self.assertEqual(parser.number_of_trials, parser_from_yaml.number_of_trials)

        self.assertEqual(parser.number_of_scores, 2)
        self.assertEqual(parser.number_of_scores, parser_from_dict.number_of_scores)
        self.assertEqual(parser.number_of_scores, parser_from_yaml.number_of_scores)

        self.assertEqual(parser.primary_score_name, 'RMSE')
        self.assertEqual(parser.primary_score_name, parser_from_dict.primary_score_name)
        self.assertEqual(parser.primary_score_name, parser_from_yaml.primary_score_name)

        self.assertTrue(all(parser.trial_rankings == grid_search_housing.cv_results_['rank_test_RMSE']))  # noqa
        self.assertEqual(parser.best_score_index, 3)
        self.assertEqual(parser.best_score,
                         grid_search_housing.cv_results_['mean_test_RMSE'][3] * -1)
        self.assertEqual(parser.best_score,
                         np.nanmin(grid_search_housing.cv_results_['mean_test_RMSE'] * -1))

        self.assertTrue(all([parser.to_dataframe().loc[parser.best_score_index, key] == value
                             for key, value in parser.best_params.items()]))

        self.assertEqual(parser.best_score, parser_from_dict.best_score)
        self.assertEqual(parser.best_score, parser_from_yaml.best_score)

    def test_MLExperimentResults_plots(self):
        new_param_column_names = {'model__max_features': 'max_features',
                                  'model__n_estimators': 'n_estimators',
                                  'model__min_samples_split': 'min_samples_split',
                                  'preparation__non_numeric_pipeline__encoder_chooser__transformer':
                                      'encoder'}
        parser = MLExperimentResults.from_sklearn_search_cv(searcher=self.credit_data__grid_search,
                                                            higher_score_is_better=True,
                                                            description="test description",
                                                            parameter_name_mappings=new_param_column_names)
        _ = parser.plot_performance_across_trials()
        _ = parser.plot_parameter_values_across_trials()
        _ = parser.plot_parallel_coordinates()
        _ = parser.plot_scatter_matrix()

        parser = MLExperimentResults.from_sklearn_search_cv(searcher=self.credit_data__grid_search__roc_auc,
                                                            higher_score_is_better=True,
                                                            description="test description",
                                                            parameter_name_mappings=None)
        _ = parser.plot_performance_across_trials()
        _ = parser.plot_parameter_values_across_trials()
        _ = parser.plot_parallel_coordinates()
        _ = parser.plot_scatter_matrix()

        parser = MLExperimentResults.from_sklearn_search_cv(searcher=self.housing_data__grid_search,
                                                            higher_score_is_better=False,
                                                            description="test description")
        _ = parser.plot_performance_across_trials()
        _ = parser.plot_parameter_values_across_trials()
        _ = parser.plot_parallel_coordinates()
        _ = parser.plot_scatter_matrix()

    def test_MLExperimentResults_gridsearch_regression_single_score(self):
        ####
        # set up grid-search on regression model for housing data
        # this will have the same search parameters as the self.housing_data__grid_search in the setUpClass,
        # except a single score (rmse, which is the same primary score as the other) rather than multiple
        # scores therefore, most of the values (e.g. best_score, best_params, should be the same between the
        # two objects)
        ####
        housing_data = get_data_housing()
        housing_data.loc[0:46, ['median_income']] = np.nan
        housing_data.loc[25:75, ['housing_median_age']] = np.nan
        y_full = housing_data['target']
        X_full = housing_data.drop(columns='target')  # noqa
        X_train, X_test, y_train, y_test = train_test_split(X_full, y_full, test_size=0.2, random_state=42)  # noqa
        del y_full, X_full
        numeric_pipeline = Pipeline([
            ('imputing', SimpleImputer(strategy='mean')),
            ('scaling', StandardScaler()),
        ])
        random_forest_model = RandomForestRegressor(random_state=42)
        full_pipeline = Pipeline([
            ('preparation', numeric_pipeline),
            ('model', random_forest_model)
        ])
        param_grad = [
            {
                'model__max_features': [2, 'auto'],
                'model__n_estimators': [10, 50]
            },
        ]
        grid_search_housing = GridSearchCV(
            full_pipeline,
            param_grid=param_grad,
            cv=RepeatedKFold(n_splits=3, n_repeats=1, random_state=42),
            scoring='neg_root_mean_squared_error',
            return_train_score=True
        )
        grid_search_housing.fit(X_train, y_train)
        results_single_score = MLExperimentResults.from_sklearn_search_cv(
            searcher=grid_search_housing,
            higher_score_is_better=False,
            description="test description"
        )
        yaml_file = get_test_path('sklearn_eval/housing_data__grid_search__single_score.yaml')
        os.remove(yaml_file)
        results_single_score.to_yaml_file(yaml_file)

        results_multi_score = MLExperimentResults.from_sklearn_search_cv(
            searcher=self.housing_data__grid_search,
            higher_score_is_better=False,
            description="test description"
        )
        self.assertEqual(results_multi_score.best_score, results_single_score.best_score)
        self.assertEqual(results_multi_score.best_params, results_single_score.best_params)
        self.assertEqual(results_multi_score.best_score_index, results_single_score.best_score_index)
        self.assertEqual(list(results_multi_score.best_trial_indexes), list(results_single_score.best_trial_indexes))  # noqa
        self.assertEqual(list(results_multi_score.trial_rankings), list(results_single_score.trial_rankings))
        self.assertEqual(results_multi_score.best_standard_error, results_single_score.best_standard_error)
        self.assertEqual(results_multi_score.parameter_names, results_single_score.parameter_names)
        self.assertEqual(results_multi_score.parameter_names_original, results_single_score.parameter_names_original)  # noqa
        self.assertEqual(results_multi_score.parameter_names_mapping, results_single_score.parameter_names_mapping)  # noqa

        self.assertEqual(results_multi_score.test_score_averages['RMSE'],
                         results_single_score.test_score_averages['neg_root_mean_squared_error'])
        self.assertEqual(results_multi_score.test_score_standard_deviations['RMSE'],
                         results_single_score.test_score_standard_deviations['neg_root_mean_squared_error'])
        self.assertEqual(results_multi_score.test_score_rankings['RMSE'],
                         results_single_score.test_score_rankings['neg_root_mean_squared_error'])

        self.assertEqual(results_multi_score.train_score_averages['RMSE'],
                         results_single_score.train_score_averages['neg_root_mean_squared_error'])
        self.assertEqual(results_multi_score.train_score_standard_deviations['RMSE'],
                         results_single_score.train_score_standard_deviations['neg_root_mean_squared_error'])

        self.assertTrue(all(results_multi_score.primary_score_averages == results_single_score.primary_score_averages))  # noqa

        self.assertEqual(results_multi_score.trial_labels(), results_single_score.trial_labels())
        self.assertEqual(results_multi_score.indexes_within_1_standard_error, results_single_score.indexes_within_1_standard_error)  # noqa
        self.assertTrue(all(results_multi_score.score_standard_errors('RMSE') == results_single_score.score_standard_errors('neg_root_mean_squared_error')))  # noqa

    def test_TwoClassEvaluator(self):
        y_true = self.credit_data__y_test
        y_score = self.credit_data__y_scores
        score_threshold = 0.5
        y_pred = [1 if x > score_threshold else 0 for x in y_score]
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      positive_class='Defaulted',
                                      negative_class='Not Defaulted',
                                      score_threshold=0.5)

        con_matrix = confusion_matrix(y_true=y_true, y_pred=y_pred)
        self.assertEqual(evaluator._true_negatives, con_matrix[0, 0])
        self.assertEqual(evaluator._false_positives, con_matrix[0, 1])
        self.assertEqual(evaluator._false_negatives, con_matrix[1, 0])
        self.assertEqual(evaluator._true_positives, con_matrix[1, 1])
        self.assertEqual(evaluator.sample_size, len(self.credit_data__y_test))
        self.assertTrue(evaluator.true_positive_rate == evaluator.sensitivity == evaluator.recall)
        self.assertEqual(evaluator.recall, recall_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(evaluator.true_negative_rate, evaluator.specificity)
        self.assertEqual(round(evaluator.false_positive_rate, 9), round(1 - evaluator.specificity, 9))
        self.assertEqual(evaluator.positive_predictive_value, evaluator.precision)
        self.assertEqual(evaluator.precision, precision_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(evaluator.f1_score, f1_score(y_true=y_true, y_pred=y_pred))
        self.assertEqual(evaluator.auc, roc_auc_score(y_true=y_true, y_score=y_score))
        self.assertEqual(evaluator.fbeta_score(beta=0), fbeta_score(y_true=y_true, y_pred=y_pred, beta=0))
        self.assertEqual(evaluator.fbeta_score(beta=1), fbeta_score(y_true=y_true, y_pred=y_pred, beta=1))
        self.assertEqual(evaluator.fbeta_score(beta=1), evaluator.f1_score)
        self.assertEqual(evaluator.fbeta_score(beta=2), fbeta_score(y_true=y_true, y_pred=y_pred, beta=2))
        self.assertEqual(round(evaluator.kappa, 9), round(cohen_kappa_score(y1=y_true, y2=y_pred), 9))

        helper_test_dataframe(file_name=get_test_path('sklearn_eval/get_auc_curve_dataframe.txt'),
                              dataframe=evaluator._get_auc_curve_dataframe())

        helper_test_dataframe(file_name=get_test_path('sklearn_eval/get_threshold_curve_dataframe.txt'),
                              dataframe=evaluator._get_threshold_curve_dataframe())

        check_plot(file_name=get_test_path('sklearn_eval/plot_predicted_scores_histogram.png'),
                   plot_function=lambda: evaluator.plot_predicted_scores_histogram())

        self.assertIsInstance(evaluator.all_metrics, dict)
        self.assertIsInstance(evaluator.all_metrics_df(return_style=False), pd.DataFrame)

        with open(get_test_path('sklearn_eval/all_metrics_df.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_explanations=False,
                                                  dummy_classifier_strategy=None,
                                                  return_style=True).to_html()
            file.write(clean_formatted_dataframe(table_html))

        with open(get_test_path('sklearn_eval/all_metrics_df__dummy.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_explanations=False,
                                                  dummy_classifier_strategy='prior',
                                                  return_style=True).to_html()
            file.write(clean_formatted_dataframe(table_html))

        with open(get_test_path('sklearn_eval/all_metrics_df__dummies.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_explanations=False,
                                                  dummy_classifier_strategy=['prior', 'constant'],
                                                  return_style=True).to_html()
            file.write(clean_formatted_dataframe(table_html))

        with open(get_test_path('sklearn_eval/all_metrics_df__round_3.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_explanations=False,
                                                  return_style=True,
                                                  round_by=3).to_html()
            file.write(clean_formatted_dataframe(table_html))

        with open(get_test_path('sklearn_eval/all_metrics_df__with_details.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_explanations=True, return_style=True).to_html()
            file.write(clean_formatted_dataframe(table_html))

        with open(get_test_path('sklearn_eval/all_metrics_df__with_details__round_3.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_explanations=True,
                                                  return_style=True,
                                                  round_by=3).to_html()
            file.write(clean_formatted_dataframe(table_html))

    def test_RegressionEvaluator(self):
        evaluator = RegressionEvaluator(actual_values=self.housing_data__y_test,
                                        predicted_values=self.housing_data__y_predictions)

        self.assertEqual(evaluator.mean_absolute_error,
                         mean_absolute_error(y_true=self.housing_data__y_test,
                                             y_pred=self.housing_data__y_predictions))
        self.assertEqual(round(mean_squared_error(y_true=self.housing_data__y_test,
                                                  y_pred=self.housing_data__y_predictions), 4),
                         round(evaluator.mean_squared_error, 4))
        self.assertEqual(np.sqrt(evaluator.mean_squared_error), evaluator.root_mean_squared_error)
        self.assertEqual(evaluator.total_observations, len(self.housing_data__y_test))
        self.assertIsInstance(evaluator.all_metrics, dict)
        self.assertIsInstance(evaluator.all_metrics_df(), pd.DataFrame)

        with open(get_test_path('sklearn_eval/reg_eval__all_metrics_df.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_style=True,
                                                  dummy_regressor_strategy=None).to_html()
            file.write(clean_formatted_dataframe(table_html))

        with open(get_test_path('sklearn_eval/reg_eval__all_metrics_df__dummy.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_style=True,
                                                  dummy_regressor_strategy='mean').to_html()
            file.write(clean_formatted_dataframe(table_html))

        with open(get_test_path('sklearn_eval/reg_eval__all_metrics_df__dummies.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_style=True,
                                                  dummy_regressor_strategy=['mean', 'median']).to_html()
            file.write(clean_formatted_dataframe(table_html))

        with open(get_test_path('sklearn_eval/reg_eval__all_metrics_df__round_3.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_style=True, round_by=3).to_html()
            file.write(clean_formatted_dataframe(table_html))

        with open(get_test_path('sklearn_eval/reg_eval__all_metrics_df__round_0.html'), 'w') as file:
            table_html = evaluator.all_metrics_df(return_style=True, round_by=0).to_html()
            file.write(clean_formatted_dataframe(table_html))

        check_plot(file_name=get_test_path('sklearn_eval/reg_eval__plot_residuals_vs_fits.png'),
                   plot_function=lambda: evaluator.plot_residuals_vs_fits())

        check_plot(file_name=get_test_path('sklearn_eval/reg_eval__plot_residuals_vs_actuals.png'),
                   plot_function=lambda: evaluator.plot_residuals_vs_actuals())

        check_plot(file_name=get_test_path('sklearn_eval/reg_eval__plot_predictions_vs_actuals.png'),
                   plot_function=lambda: evaluator.plot_predictions_vs_actuals())

    def test_plot_confusion_matrix(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      positive_class='Defaulted',
                                      negative_class='Not Defaulted',
                                      score_threshold=0.5)

        check_plot(file_name=get_test_path('sklearn_eval/plot_confusion_matrix.png'),
                   plot_function=lambda: evaluator.plot_confusion_matrix())

    def test_plot_roc_auc_curve(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      positive_class='Defaulted',
                                      negative_class='Not Defaulted',
                                      score_threshold=0.5)

        check_plot(file_name=get_test_path('sklearn_eval/plot_auc_curve.png'),
                   plot_function=lambda: evaluator.plot_roc_auc_curve(return_plotly=False))

    def test_plot_precision_recall_auc_curve(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      positive_class='Defaulted',
                                      negative_class='Not Defaulted',
                                      score_threshold=0.5)

        check_plot(file_name=get_test_path('sklearn_eval/plot_precision_recall_auc_curve.png'),
                   plot_function=lambda: evaluator.plot_precision_recall_auc_curve(return_plotly=False))

    def test_plot_threshold_curves(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      positive_class='Defaulted',
                                      negative_class='Not Defaulted',
                                      score_threshold=0.5)

        check_plot(file_name=get_test_path('sklearn_eval/plot_threshold_curves.png'),
                   plot_function=lambda: evaluator.plot_threshold_curves(return_plotly=False))

    def test_plot_precision_recall_tradeoff(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      positive_class='Defaulted',
                                      negative_class='Not Defaulted',
                                      score_threshold=0.5)

        check_plot(file_name=get_test_path('sklearn_eval/plot_precision_recall_tradeoff.png'),
                   plot_function=lambda: evaluator.plot_precision_recall_tradeoff(return_plotly=False))

    def test_calculate_lift_gain(self):
        evaluator = TwoClassEvaluator(actual_values=self.credit_data__y_test,
                                      predicted_scores=self.credit_data__y_scores,
                                      positive_class='Defaulted',
                                      negative_class='Not Defaulted',
                                      score_threshold=0.5)

        helper_test_dataframe(file_name=get_test_path('sklearn_eval/calculate_lift_gain.txt'),
                              dataframe=evaluator.calculate_lift_gain())

        helper_test_dataframe(file_name=get_test_path('sklearn_eval/calculate_lift_gain__10_buckets.txt'),
                              dataframe=evaluator.calculate_lift_gain(num_buckets=10))

        with open(get_test_path('sklearn_eval/calculate_lift_gain.html'), 'w') as file:
            table_html = evaluator.calculate_lift_gain(return_style=True).to_html()
            file.write(clean_formatted_dataframe(table_html))

        with open(get_test_path('sklearn_eval/calculate_lift_gain__10_buckets.html'), 'w') as file:
            table_html = evaluator.calculate_lift_gain(return_style=True, num_buckets=10).to_html()
            file.write(clean_formatted_dataframe(table_html))


if __name__ == '__main__':
    unittest.main()



import datetime
import unittest
from contextlib import contextmanager
from enum import unique, Enum, auto
from os import path
from unittest.mock import patch

import numpy as np
import pandas as pd

from helpsk.database import GenericConfigFile, RedshiftConfigFile, SnowflakeConfigFile, Redshift, Snowflake
from tests.helpers import get_test_path


@unique
class TestEnum(Enum):
    VALUE_A = auto()
    VALUE_B = auto()


@contextmanager
def mock_redshift():
    """This is basically a wrapper around mock.patch to hide the mock logic. i.e. This is a context manager
    that wraps the mock context wrapper
    """
    with patch('psycopg2.connect'), patch('pandas.read_sql_query') as pandas_read_mock:
        pandas_read_mock.return_value = pd.DataFrame({'test': ['test']})
        yield


@contextmanager
def mock_snowflake():
    """This is basically a wrapper around mock.patch to hide the mock logic. i.e. This is a context manager
    that wraps the mock context wrapper
    """
    with patch('snowflake.connector.connect') as mock_snowflake_connector:
        # mock snowflake logic:
        # connect() -> cursor() -> ... -> fetch_pandas_all()
        # actual code
        # self.connection_object = snowflake.connector.connect()
        # cursor = self.connection_object.cursor()
        # cursor.execute(sql)
        # dataframe = cursor.fetch_pandas_all()
        mock_con = mock_snowflake_connector.return_value
        mock_cur = mock_con.cursor.return_value
        mock_cur.fetch_pandas_all.return_value = pd.DataFrame({'test': ['test']})
        yield


# noinspection PyMethodMayBeStatic
class TestDatabase(unittest.TestCase):

    def setUp(self) -> None:
        self.sample_redshift_file = get_test_path('database/sample_redshift.config')
        self.assertTrue(path.isfile(self.sample_redshift_file))

        self.sample_snowflake_file = get_test_path('database/sample_snowflake.config')
        self.assertTrue(path.isfile(self.sample_snowflake_file))

        self.sample_snowflake_autocommit_true_file = get_test_path('database/sample_snowflake_autocommit_true.config')  # noqa
        self.assertTrue(path.isfile(self.sample_snowflake_autocommit_true_file))

        self.sample_snowflake_autocommit_false_file = get_test_path('database/sample_snowflake_autocommit_false.config')  # noqa
        self.assertTrue(path.isfile(self.sample_snowflake_autocommit_false_file))

    def test_GenericConfigFile(self):
        config_mapping = {'user_param': 'username',
                          'password_param': 'password',
                          'database_param': 'database',
                          'port_param': 'port',
                          'host_param': 'host'}
        config = GenericConfigFile(config_file=self.sample_redshift_file,
                                   config_key='redshift',
                                   config_mapping=config_mapping)
        expected_value = {'user_param': 'my_username',
                          'password_param': 'my-password-123',
                          'database_param': 'the_database',
                          'port_param': '1234',
                          'host_param': 'host.address.redshift.amazonaws.com'}
        self.assertEqual(config.get_dictionary(), expected_value)

    def test_RedshiftConfigFile(self):
        config = RedshiftConfigFile(config_file=self.sample_redshift_file)
        config_dict = config.get_dictionary()
        expected_value = {'user': 'my_username',
                          'password': 'my-password-123',
                          'database': 'the_database',
                          'port': '1234',
                          'host': 'host.address.redshift.amazonaws.com'}
        self.assertEqual(expected_value, config_dict)

        # test that this can be passed into the redshift database object, which implies that config_dict
        # contains the proper keyboard arguments
        Redshift(**config_dict)

    def test_SnowflakeConfigFile(self):
        config = SnowflakeConfigFile(config_file=self.sample_snowflake_file)
        config_dict = config.get_dictionary()
        expected_value = {
            'user': 'my.email@address.com',
            'account': 'account.id',
            'authenticator': 'externalbrowser',
            'warehouse': 'WAREHOUSE_NAME',
            'database': 'DATABASE_NAME',
        }
        self.assertEqual(expected_value, config_dict)

        # test that this can be passed into the snowflake database object, which implies that config_dict
        # contains the proper keyboard arguments
        Snowflake(**config_dict)

    def test_SnowflakeConfigFile_auto_commit(self):
        config = SnowflakeConfigFile(config_file=self.sample_snowflake_file)
        snowflake = Snowflake.from_config(config=config)
        self.assertIsInstance(snowflake.autocommit, bool)
        self.assertTrue(snowflake.autocommit)

        config = SnowflakeConfigFile(config_file=self.sample_snowflake_autocommit_false_file)
        self.assertTrue('autocommit' in config.get_dictionary())
        self.assertEqual(config.get_dictionary()['autocommit'], 'FALSE')
        snowflake = Snowflake.from_config(config=config)
        self.assertIsInstance(snowflake.autocommit, bool)
        self.assertFalse(snowflake.autocommit)

        config = SnowflakeConfigFile(config_file=self.sample_snowflake_autocommit_true_file)
        self.assertTrue('autocommit' in config.get_dictionary())
        self.assertEqual(config.get_dictionary()['autocommit'], 'TRUE')
        snowflake = Snowflake.from_config(config=config)
        self.assertIsInstance(snowflake.autocommit, bool)
        self.assertTrue(snowflake.autocommit)

    def test_Snowflake_generate_sql_create_table(self):
        sample_data = pd.DataFrame({'col_a': [np.nan, 2, 3, 4],
                                    'col_b': [np.nan, 'b', 'd', 'd'],
                                    'col_bb': ['a', 'b', 'd', 'd'],
                                    'col_c': pd.date_range('2021-01-01', '2021-01-04'),
                                    'col_d': [datetime.date(2021, 4, 2), datetime.date(2021, 4, 2),
                                              datetime.date(2021, 4, 2), datetime.date(2021, 4, 2)],
                                    'col_e': np.nan,
                                    'col_f': [1.0, 2.0, 3.0, 4.0],
                                    'col_h': [TestEnum.VALUE_A, TestEnum.VALUE_A,
                                              TestEnum.VALUE_B, TestEnum.VALUE_B],
                                    'col_j': [False, False, True, False],
                                    'col_k': [None, None, None, None],
                                    'col_l': [np.nan, np.nan, np.nan, np.nan]
                                    })
        sample_data['col_g'] = sample_data['col_b'].astype('category')

        sql = Snowflake._generate_sql_create_table(
            table='table_name',
            dataframe=sample_data,
            database='database_name',
            schema='schema_name',
        )
        expected_value = """CREATE OR REPLACE TABLE DATABASE_NAME.SCHEMA_NAME.TABLE_NAME (
            col_a float8,
            col_b varchar,
            col_bb varchar,
            col_c datetime,
            col_d varchar,
            col_e float8,
            col_f float8,
            col_h varchar,
            col_j boolean,
            col_k varchar,
            col_l float8,
            col_g varchar
        )""".replace("        ", "")
        self.assertEqual(sql, expected_value)

        sql = Snowflake._generate_sql_create_table(
            table='table_name',
            dataframe=sample_data,
            database='database_name',
            # schema='schema_name',
        )
        expected_value = """CREATE OR REPLACE TABLE DATABASE_NAME.TABLE_NAME (
            col_a float8,
            col_b varchar,
            col_bb varchar,
            col_c datetime,
            col_d varchar,
            col_e float8,
            col_f float8,
            col_h varchar,
            col_j boolean,
            col_k varchar,
            col_l float8,
            col_g varchar
        )""".replace("        ", "")
        self.assertEqual(sql, expected_value)

        sql = Snowflake._generate_sql_create_table(
            table='table_name',
            dataframe=sample_data,
            # database='database_name',
            schema='schema_name',
        )
        expected_value = """CREATE OR REPLACE TABLE SCHEMA_NAME.TABLE_NAME (
                    col_a float8,
                    col_b varchar,
                    col_bb varchar,
                    col_c datetime,
                    col_d varchar,
                    col_e float8,
                    col_f float8,
                    col_h varchar,
                    col_j boolean,
                    col_k varchar,
                    col_l float8,
                    col_g varchar
                )""".replace("        ", "")
        self.assertEqual(sql, expected_value)

        sql = Snowflake._generate_sql_create_table(
            table='table_name',
            dataframe=sample_data,
            # database='database_name',
            # schema='schema_name',
        )
        expected_value = """CREATE OR REPLACE TABLE TABLE_NAME (
                            col_a float8,
                            col_b varchar,
                            col_bb varchar,
                            col_c datetime,
                            col_d varchar,
                            col_e float8,
                            col_f float8,
                            col_h varchar,
                            col_j boolean,
                            col_k varchar,
                            col_l float8,
                            col_g varchar
                        )""".replace("        ", "")
        self.assertEqual(sql, expected_value)

    def test_Database(self):
        redshift_config = RedshiftConfigFile(config_file=self.sample_redshift_file)
        snowflake_config = SnowflakeConfigFile(config_file=self.sample_snowflake_file)

        def test_via_constructor(db_obj, db_mock):
            self.assertFalse(db_obj.is_connected())
            self.assertIsNone(db_obj.connection_object)

            # mock connection method so that we can "open" the connection to the database
            with db_mock():
                """db_obj is a Database object that is in a closed state and has the mock objects set up"""
                for _ in range(2):  # test that the same object can be opened/closed multiple times
                    # connection should be closed
                    self.assertFalse(db_obj.is_connected())
                    self.assertIsNone(db_obj.connection_object)

                    # test connecting
                    db_obj.connect()
                    self.assertTrue(db_obj.is_connected())
                    self.assertIsNotNone(db_obj.connection_object)
                    self.assertIsInstance(db_obj.connection_object, unittest.mock.MagicMock)
                    db_obj.connect()  # test that calling open again doesn't not fail

                    # test querying
                    results = db_obj.query("SELECT * FROM doesnt_exist LIMIT 100", show_elapsed_time=False)
                    self.assertIsInstance(results, pd.DataFrame)
                    self.assertEqual(results.iloc[0, 0], 'test')
                    # test connection is still open after querying
                    self.assertTrue(db_obj.is_connected())
                    self.assertIsNotNone(db_obj.connection_object)
                    # test subsequent query
                    results = db_obj.query("SELECT * FROM doesnt_exist LIMIT 100", show_elapsed_time=False)
                    self.assertIsInstance(results, pd.DataFrame)
                    self.assertEqual(results.iloc[0, 0], 'test')

                    # test closing
                    db_obj.close()
                    self.assertFalse(db_obj.is_connected())
                    self.assertIsNone(db_obj.connection_object)

        test_via_constructor(db_obj=Redshift(**redshift_config.get_dictionary()), db_mock=mock_redshift)
        test_via_constructor(db_obj=Redshift.from_config(redshift_config), db_mock=mock_redshift)
        test_via_constructor(db_obj=Snowflake(**snowflake_config.get_dictionary()), db_mock=mock_snowflake)
        test_via_constructor(db_obj=Snowflake.from_config(snowflake_config), db_mock=mock_snowflake)

        def test_context_manager(db_class, db_config, db_mock):
            # test context manager
            with db_mock():
                with db_class.from_config(db_config) as db_object:
                    self.assertTrue(db_object.is_connected())
                    self.assertIsNotNone(db_object.connection_object)
                    self.assertIsInstance(db_object.connection_object, unittest.mock.MagicMock)

                    # test querying
                    results = db_object.query("SELECT * FROM doesnt_exist LIMIT 100", show_elapsed_time=False)
                    self.assertIsInstance(results, pd.DataFrame)
                    self.assertEqual(results.iloc[0, 0], 'test')
                    # test connection is still open after querying
                    self.assertTrue(db_object.is_connected())
                    self.assertIsNotNone(db_object.connection_object)

                self.assertFalse(db_object.is_connected())
                self.assertIsNone(db_object.connection_object)

        test_context_manager(db_class=Redshift, db_config=redshift_config, db_mock=mock_redshift)
        test_context_manager(db_class=Snowflake, db_config=snowflake_config, db_mock=mock_snowflake)

        # tests that if failure to connect (i.e. no mock and connection failure) that the object is not in an
        # open state
        database = Redshift.from_config(redshift_config)
        with self.assertRaises(Exception):
            database.connect()
        self.assertFalse(database.is_connected())
        self.assertIsNone(database.connection_object)


if __name__ == '__main__':
    unittest.main()



"""This module contains helper functions when working with sklearn (scikit-learn) objects;
in particular, for building pipelines"""

from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import OrdinalEncoder


class TransformerChooser(BaseEstimator, TransformerMixin):
    """Transformer that wraps another Transformer. This allows different transformer objects to be tuned.
    """
    def __init__(self, transformer=None):
        """
        Args:
            transformer:
                Transformer object (e.g. StandardScaler, MinMaxScaler)
        """
        self.transformer = transformer

    def fit(self, X, y=None):
        """fit implementation
        """
        if self.transformer is None:
            return self

        return self.transformer.fit(X, y)

    def transform(self, X):
        """transform implementation
        """
        if self.transformer is None:
            return X

        return self.transformer.transform(X)


class CustomOrdinalEncoder(BaseEstimator, TransformerMixin):
    """First replaces missing values with '<missing>' then applies OrdinalEncoder
    """

    def __init__(self, handle_unknown='use_encoded_value'):
        self._ordinal_encoder = OrdinalEncoder(handle_unknown=handle_unknown, unknown_value=-1)
        # unknown_value=-1,
        # handle_unknown='use_encoded_value')
        self.handle_unknown = handle_unknown
        self._missing_value = '<missing>'

    def _fill_na(self, X):
        """Helper function that fills missing values with strings before calling OrdinalEncoder"""
        for column in X.columns.values:
            if X[column].dtype.name == 'category':
                if self._missing_value not in X[column].cat.categories:
                    X[column] = X[column].cat.add_categories(self._missing_value)
                X[column] = X[column].fillna(self._missing_value)

        return X

    def fit(self, X, y=None):
        """fit implementation"""
        X = self._fill_na(X)
        self._ordinal_encoder.fit(X)
        return self

    def transform(self, X):
        """transform implementation"""
        X = self._fill_na(X)
        return self._ordinal_encoder.transform(X)



import unittest
from datetime import timedelta

import numpy as np
import pandas as pd

from helpsk import string as hs
from helpsk import validation as hv
from helpsk import exceptions as he


# noinspection PyMethodMayBeStatic
class TestValidation(unittest.TestCase):

    def test_any_none_nan(self):
        self.assertTrue(hv.any_none_nan(None))
        self.assertTrue(hv.any_none_nan(np.NaN))
        self.assertTrue(hv.any_none_nan(pd.NaT))
        self.assertTrue(hv.any_none_nan(pd.NA))

        self.assertFalse(hv.any_none_nan(0))
        self.assertFalse(hv.any_none_nan(1))
        self.assertFalse(hv.any_none_nan(''))
        self.assertFalse(hv.any_none_nan('a'))

        # test list
        self.assertTrue(hv.any_none_nan([1, np.nan, None]))
        self.assertTrue(hv.any_none_nan([1, np.nan]))
        self.assertTrue(hv.any_none_nan([1, pd.NA]))
        self.assertTrue(hv.any_none_nan([1, pd.NaT]))
        self.assertTrue(hv.any_none_nan([1, None]))
        self.assertTrue(hv.any_none_nan([np.nan]))
        self.assertTrue(hv.any_none_nan([pd.NA]))
        self.assertTrue(hv.any_none_nan([pd.NaT]))
        self.assertTrue(hv.any_none_nan([None]))
        self.assertTrue(hv.any_none_nan([]))
        self.assertFalse(hv.any_none_nan([1]))
        self.assertFalse(hv.any_none_nan(['']))

        # test numpy array
        self.assertTrue(hv.any_none_nan(np.array([1, np.nan, None])))
        self.assertTrue(hv.any_none_nan(np.array([1, np.nan])))
        self.assertTrue(hv.any_none_nan(np.array([1, pd.NA])))
        self.assertTrue(hv.any_none_nan(np.array([1, pd.NaT])))
        self.assertTrue(hv.any_none_nan(np.array([1, None])))
        self.assertTrue(hv.any_none_nan(np.array([np.nan])))
        self.assertTrue(hv.any_none_nan(np.array([pd.NA])))
        self.assertTrue(hv.any_none_nan(np.array([pd.NaT])))
        self.assertTrue(hv.any_none_nan(np.array([None])))
        self.assertTrue(hv.any_none_nan(np.array([])))
        self.assertFalse(hv.any_none_nan(np.array([1])))
        self.assertFalse(hv.any_none_nan(np.array([''])))

        # test pandas series
        self.assertTrue(hv.any_none_nan(pd.Series([1, np.nan, None])))
        self.assertTrue(hv.any_none_nan(pd.Series([1, np.nan])))
        self.assertTrue(hv.any_none_nan(pd.Series([1, pd.NA])))
        self.assertTrue(hv.any_none_nan(pd.Series([1, pd.NaT])))
        self.assertTrue(hv.any_none_nan(pd.Series([1, None])))
        self.assertTrue(hv.any_none_nan(pd.Series([np.nan])))
        self.assertTrue(hv.any_none_nan(pd.Series([pd.NA])))
        self.assertTrue(hv.any_none_nan(pd.Series([pd.NaT])))
        self.assertTrue(hv.any_none_nan(pd.Series([None])))
        self.assertTrue(hv.any_none_nan(pd.Series([], dtype=float)))
        self.assertFalse(hv.any_none_nan(pd.Series([1])))
        self.assertFalse(hv.any_none_nan(pd.Series([''])))

        # test pandas data.frame
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[1, np.nan, None], [1, 2, 3]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[1, np.nan], [1, 2]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[1, pd.NA], [1, 2]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[1, pd.NaT], [1, 2]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[1, None], [1, 2]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[np.nan], [1]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[pd.NA], [1]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[pd.NaT], [1]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([[None], [1]])))
        self.assertTrue(hv.any_none_nan(pd.DataFrame([], dtype=float)))
        self.assertFalse(hv.any_none_nan(pd.DataFrame([1])))
        self.assertFalse(hv.any_none_nan(pd.DataFrame([[1], [1]])))
        self.assertFalse(hv.any_none_nan(pd.DataFrame([[''], [1]])))

    def test_assert_not_none_nan(self):

        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(None)
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(np.NaN)
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.NA)
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.NaT)

        hv.assert_not_none_nan(0)
        hv.assert_not_none_nan('')

        # test list
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan([1, np.nan, None])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan([1, np.nan])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan([1, None])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan([np.nan])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan([None])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan([])
        hv.assert_not_none_nan([1])
        hv.assert_not_none_nan([''])

        # test numpy array
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(np.array([1, np.nan, None]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(np.array([1, np.nan]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(np.array([1, None]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(np.array([np.nan]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(np.array([None]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(np.array([]))
        hv.assert_not_none_nan(np.array([1]))
        hv.assert_not_none_nan(np.array(['']))

        # test pandas series
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.Series([1, np.nan, None]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.Series([1, np.nan]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.Series([1, None]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.Series([np.nan]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.Series([None]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.Series([], dtype=float))
        hv.assert_not_none_nan(pd.Series([1]))
        hv.assert_not_none_nan(pd.Series(['']))

        # test pandas data.frame
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.DataFrame([[1, np.nan, None], [1, 2, 3]]))  # noqa
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.DataFrame([[1, np.nan], [1, 2]]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.DataFrame([[1, None], [1, 2]]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.DataFrame([[np.nan], [1]]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.DataFrame([[None], [1]]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_none_nan(pd.DataFrame([], dtype=float))
        hv.assert_not_none_nan(pd.DataFrame([1]))
        hv.assert_not_none_nan(pd.DataFrame([[1], [1]]))
        hv.assert_not_none_nan(pd.DataFrame([[''], [1]]))

    def test_any_missing(self):
        self.assertTrue(hv.any_missing(None))
        self.assertTrue(hv.any_missing(np.NaN))
        self.assertTrue(hv.any_missing(pd.NA))
        self.assertTrue(hv.any_missing(pd.NaT))
        self.assertTrue(hv.any_missing(''))

        self.assertFalse(hv.any_missing(0))

        # test list
        self.assertTrue(hv.any_missing([1, np.nan, None]))
        self.assertTrue(hv.any_missing([1, np.nan]))
        self.assertTrue(hv.any_missing([1, pd.NA]))
        self.assertTrue(hv.any_missing([1, pd.NaT]))
        self.assertTrue(hv.any_missing([1, None]))
        self.assertTrue(hv.any_missing([np.nan]))
        self.assertTrue(hv.any_missing([pd.NA]))
        self.assertTrue(hv.any_missing([pd.NaT]))
        self.assertTrue(hv.any_missing([None]))
        self.assertTrue(hv.any_missing([]))
        self.assertTrue(hv.any_missing(['']))
        self.assertTrue(hv.any_missing(['abc', '']))
        self.assertTrue(hv.any_missing([1, '']))
        self.assertFalse(hv.any_missing([1]))
        self.assertFalse(hv.any_missing(['a']))

        # test pandas series
        self.assertTrue(hv.any_missing(pd.Series([1, np.nan, None])))
        self.assertTrue(hv.any_missing(pd.Series([1, np.nan])))
        self.assertTrue(hv.any_missing(pd.Series([1, pd.NA])))
        self.assertTrue(hv.any_missing(pd.Series([1, pd.NaT])))
        self.assertTrue(hv.any_missing(pd.Series([1, None])))
        self.assertTrue(hv.any_missing(pd.Series([np.nan])))
        self.assertTrue(hv.any_missing(pd.Series([pd.NA])))
        self.assertTrue(hv.any_missing(pd.Series([pd.NaT])))
        self.assertTrue(hv.any_missing(pd.Series([None])))
        self.assertTrue(hv.any_missing(pd.Series([], dtype=float)))
        self.assertTrue(hv.any_missing(pd.Series([''])))
        self.assertTrue(hv.any_missing(pd.Series(['abc', ''])))
        self.assertTrue(hv.any_missing(pd.Series([1, ''])))
        self.assertFalse(hv.any_missing(pd.Series([1])))
        self.assertFalse(hv.any_missing(pd.Series(['a'])))

        # test pandas data.frame
        self.assertTrue(hv.any_missing(pd.DataFrame([[1, np.nan, None], [1, 2, 3]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[1, np.nan], [1, 2]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[1, pd.NA], [1, 2]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[1, pd.NaT], [1, 2]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[1, None], [1, 2]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[np.nan], [1]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[pd.NA], [1]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[pd.NaT], [1]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([[None], [1]])))
        self.assertTrue(hv.any_missing(pd.DataFrame([], dtype=float)))
        self.assertTrue(hv.any_missing(pd.DataFrame([['abc', ''], ['abc', 'abc']])))
        self.assertTrue(hv.any_missing(pd.DataFrame([''])))
        self.assertFalse(hv.any_missing(pd.DataFrame([1])))
        self.assertFalse(hv.any_missing(pd.DataFrame([[1], [1]])))

    def test_assert_not_any_missing(self):
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(None)
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(np.NaN)
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.NA)
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.NaT)
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing('')

        hv.assert_not_any_missing(0)

        # test list
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing([1, np.nan, None])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing([1, np.nan])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing([1, None])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing([np.nan])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing([None])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing([])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing([''])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(['abc', ''])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing([1, ''])
        hv.assert_not_any_missing([1])
        hv.assert_not_any_missing(['a'])

        # test pandas series
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.Series([1, np.nan, None]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.Series([1, np.nan]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.Series([1, None]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.Series([np.nan]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.Series([None]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.Series([], dtype=float))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.Series(['']))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.Series(['abc', '']))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.Series([1, '']))
        hv.assert_not_any_missing(pd.Series([1]))
        hv.assert_not_any_missing(pd.Series(['a']))

        # test pandas data.frame
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.DataFrame([[1, np.nan, None], [1, 2, 3]]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.DataFrame([[1, np.nan], [1, 2]]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.DataFrame([[1, None], [1, 2]]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.DataFrame([[np.nan], [1]]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.DataFrame([[None], [1]]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.DataFrame([], dtype=float))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.DataFrame([['abc', ''], ['abc', 'abc']]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_any_missing(pd.DataFrame(['']))
        hv.assert_not_any_missing(pd.DataFrame([1]))
        hv.assert_not_any_missing(pd.DataFrame([[1], [1]]))

    def test_any_duplicated(self):
        # test list
        self.assertFalse(hv.any_duplicated(['']))
        self.assertFalse(hv.any_duplicated(['', 1]))
        self.assertFalse(hv.any_duplicated(['', 1, None]))

        self.assertTrue(hv.any_duplicated(['', 1, '']))
        self.assertTrue(hv.any_duplicated(['', 1, 1]))
        self.assertTrue(hv.any_duplicated(['', 1, None, None]))

        # test pd.Series
        self.assertFalse(hv.any_duplicated(pd.Series([''])))
        self.assertFalse(hv.any_duplicated(pd.Series(['', 1])))
        self.assertFalse(hv.any_duplicated(pd.Series(['', 1, None])))

        self.assertTrue(hv.any_duplicated(pd.Series(['', 1, ''])))
        self.assertTrue(hv.any_duplicated(pd.Series(['', 1, 1])))
        self.assertTrue(hv.any_duplicated(pd.Series(['', 1, None, None])))

    def test_assert_not_duplicated(self):
        # test list
        hv.assert_not_duplicated([''])
        hv.assert_not_duplicated(['', 1])
        hv.assert_not_duplicated(['', 1, None])

        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_duplicated(['', 1, ''])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_duplicated(['', 1, 1])
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_duplicated(['', 1, None, None])

        # test pd.Series
        hv.assert_not_duplicated(pd.Series(['']))
        hv.assert_not_duplicated(pd.Series(['', 1]))
        hv.assert_not_duplicated(pd.Series(['', 1, None]))

        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_duplicated(pd.Series(['', 1, '']))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_duplicated(pd.Series(['', 1, 1]))
        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_not_duplicated(pd.Series(['', 1, None, None]))

    def test_assert_true(self):
        hv.assert_true(True)
        self.assertTrue(isinstance(np.bool_(True), np.bool_))
        hv.assert_true(np.bool_(True))

        with self.assertRaises(he.HelpskParamTypeError):
            hv.assert_true([])  # noqa
        with self.assertRaises(he.HelpskParamTypeError):
            hv.assert_true([True])  # noqa

        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_true(False)

        with self.assertRaises(he.HelpskAssertionError) as cm:
            hv.assert_true(False, message='my message')
        self.assertEqual(cm.exception.args[0], 'my message')

    def test_assert_false(self):
        hv.assert_false(False)
        self.assertTrue(isinstance(np.bool_(False), np.bool_))
        hv.assert_false(np.bool_(False))

        with self.assertRaises(he.HelpskParamTypeError):
            hv.assert_false([])  # noqa
        with self.assertRaises(he.HelpskParamTypeError):
            hv.assert_false([False])  # noqa

        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_false(True)

        with self.assertRaises(he.HelpskAssertionError) as cm:
            hv.assert_false(True, message='my message')
        self.assertEqual(cm.exception.args[0], 'my message')

    def test_assert_all(self):
        cases = [
            [True],
            [True, True],
            np.array([True]),
            np.array([True, True]),
            pd.Series([True]),
            pd.Series([True, True]),
            pd.DataFrame([True]),
            pd.DataFrame([True, True]),
            pd.DataFrame([[True, True], [True, True]]),
        ]
        # running assert all on these cases should not result in any assertion errors
        for case in cases:
            hv.assert_all(case)

        cases = [
            [False],
            [True, False],
            np.array([False]),
            np.array([True, False]),
            pd.Series([False]),
            pd.Series([True, False]),
            pd.DataFrame([False]),
            pd.DataFrame([True, False]),
            pd.DataFrame([[True, True], [True, False]]),
        ]
        for index, case in enumerate(cases):
            with self.subTest(index=index):
                with self.assertRaises(he.HelpskAssertionError):
                    hv.assert_all(case)

    def test_assert_not_any(self):
        cases = [
            [False],
            [False, False],
            np.array([False]),
            np.array([False, False]),
            pd.Series([False]),
            pd.Series([False, False]),
            pd.DataFrame([False]),
            pd.DataFrame([False, False]),
            pd.DataFrame([[False, False], [False, False]]),
        ]
        # running assert all on these cases should not result in any assertion errors
        for case in cases:
            hv.assert_not_any(case)

        cases = [
            [True],
            [True, False],
            np.array([True]),
            np.array([True, False]),
            pd.Series([True]),
            pd.Series([True, False]),
            pd.DataFrame([True]),
            pd.DataFrame([True, False]),
            pd.DataFrame([[False, False], [False, True]]),
        ]
        for index, case in enumerate(cases):
            with self.subTest(index=index):
                with self.assertRaises(he.HelpskAssertionError):
                    hv.assert_not_any(case)

    def test_iterables_are_equal(self):

        def subtest_iterables_are_equal(iterable_a, iterable_b, dtype):
            assert hv.iterables_are_equal(iterable_a, iterable_b)
            if dtype != str:
                assert hv.iterables_are_equal(np.array(iterable_a), np.array(iterable_b))
                assert hv.iterables_are_equal(np.array(iterable_a), iterable_b)
            assert hv.iterables_are_equal(
                pd.Series(iterable_a, dtype=dtype),
                pd.Series(iterable_b, dtype=dtype)
            )

        subtest_iterables_are_equal([], [], dtype=np.float64)
        subtest_iterables_are_equal([np.nan], [np.nan], np.float64)
        subtest_iterables_are_equal([1, 2, 3, 4], [1, 2, 3, 4], np.float64)
        subtest_iterables_are_equal([np.nan, 1, 2, 3, 4], [np.nan, 1, 2, 3, 4], np.float64)
        subtest_iterables_are_equal([np.nan, 1, 2, 3, 4, np.nan], [np.nan, 1, 2, 3, 4, np.nan], np.float64)

        subtest_iterables_are_equal(iterable_a=['a', 'b', 'c', 'd'], iterable_b=['a', 'b', 'c', 'd'], dtype=str)  # noqa
        subtest_iterables_are_equal([np.nan, 'a', 'b', 'c', 'd'], [np.nan, 'a', 'b', 'c', 'd'], str)
        subtest_iterables_are_equal([np.nan, 'a', 'b', 'c', 'd', np.nan], [np.nan, 'a', 'b', 'c', 'd', np.nan], str)  # noqa

        subtest_iterables_are_equal(['a', 'b', 'c', 4], ['a', 'b', 'c', 4], str)
        subtest_iterables_are_equal([np.nan, 'a', 'b', 'c', 4], [np.nan, 'a', 'b', 'c', 4], str)
        subtest_iterables_are_equal([np.nan, 'a', 'b', 'c', 4, np.nan], [np.nan, 'a', 'b', 'c', 4, np.nan], str)  # noqa

        def subtest_iterables_are_not_equal(iterable_a, iterable_b, dtype):
            assert not hv.iterables_are_equal(iterable_a, iterable_b)
            assert not hv.iterables_are_equal(np.array(iterable_a), np.array(iterable_b))
            assert not hv.iterables_are_equal(pd.Series(iterable_a, dtype=dtype), pd.Series(iterable_b, dtype=dtype))  # noqa

        subtest_iterables_are_not_equal(iterable_a=[], iterable_b=[1], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[1], iterable_b=[], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[], iterable_b=[np.nan], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[np.nan], iterable_b=[], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[], iterable_b=[''], dtype=str)
        subtest_iterables_are_not_equal(iterable_a=[''], iterable_b=[], dtype=str)

        subtest_iterables_are_not_equal(iterable_a=[1, 2, 3], iterable_b=[1, 2, 3, 4], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[0, 1, 2, 3], iterable_b=[1, 2, 3], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[1.000001, 2, 3], iterable_b=[1, 2, 3], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[1, 2, 3], iterable_b=[1.000001, 2, 3], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[1, 2, 3], iterable_b=[1.000001, 2, 3], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[1, 2, 3], iterable_b=[np.nan, 2, 3], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[np.nan, 2, 3], iterable_b=[1, 2, 3], dtype=np.float64)
        subtest_iterables_are_not_equal(iterable_a=[1, 2, 3], iterable_b=[None, 2, 3], dtype=np.float64)

    def test_iterables_are_equal__ordered(self):
        # found a bug where two pd.Categorical series will not return True even if they have the same values,
        # but with different ordered; but we only care about the values
        values = [np.nan, 'a', 'b', 'c', 'd']
        series = pd.Series(values)
        unordered_categorical = pd.Categorical(values=values, ordered=False)
        ordered_categorical = pd.Categorical(values=values, categories=['d', 'c', 'a', 'b'], ordered=True)

        self.assertTrue(hv.iterables_are_equal(values, series))
        self.assertTrue(hv.iterables_are_equal(values, unordered_categorical))
        self.assertTrue(hv.iterables_are_equal(values, ordered_categorical))
        self.assertTrue(hv.iterables_are_equal(series, unordered_categorical))
        self.assertTrue(hv.iterables_are_equal(series, ordered_categorical))
        self.assertTrue(hv.iterables_are_equal(unordered_categorical, ordered_categorical))

        different_values = values.copy()
        different_values[-1] = 'z'

        self.assertFalse(hv.iterables_are_equal(different_values, series))
        self.assertFalse(hv.iterables_are_equal(different_values, unordered_categorical))
        self.assertFalse(hv.iterables_are_equal(different_values, ordered_categorical))
        self.assertFalse(hv.iterables_are_equal(pd.Categorical(different_values, ordered=False),
                                                unordered_categorical))
        self.assertFalse(hv.iterables_are_equal(pd.Categorical(different_values, ordered=True),
                                                ordered_categorical))

    def test_dataframes_match(self):

        dataframe_1 = pd.DataFrame({
            'col_floats': [1.123456789, 2.123456789, 3.123456789, np.nan],
            'col_strings': [np.nan, 'a', 'b', 'c'],
            'col_enums': [np.nan, hs.RoundTo.NONE, hs.RoundTo.AUTO, hs.RoundTo.THOUSANDS],
            'col_dates': pd.date_range('2021-01-01', periods=4),
            'col_missing': [np.nan, np.nan, np.nan, np.nan]}
        )
        dataframe_1.loc[0, 'col_dates'] = np.nan

        with self.assertRaises(he.HelpskParamTypeError):
            hv.dataframes_match(dataframes=None)  # noqa

        with self.assertRaises(he.HelpskParamTypeError):
            hv.dataframes_match(dataframes=dataframe_1)  # noqa

        with self.assertRaises(he.HelpskParamValueError):
            hv.dataframes_match(dataframes=[])

        with self.assertRaises(he.HelpskParamValueError):
            hv.dataframes_match(dataframes=[dataframe_1])

        # test that there are no side effects; e.g. we set the index/column values if we ignore them
        dataframe_1_original = dataframe_1.copy()
        dataframe_2 = dataframe_1.copy()
        dataframe_2.index = ['a', 'b', 'c', 'd']
        dataframe_2.columns = ['a', 'b', 'c', 'd', 'e']
        dataframe_2_original = dataframe_2.copy()
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2],
                                            float_tolerance=1,
                                            ignore_indexes=True,
                                            ignore_column_names=True))
        self.assertTrue(dataframe_1.equals(dataframe_1_original))
        self.assertTrue((dataframe_1.index == dataframe_1_original.index).all())
        self.assertTrue((dataframe_1.columns == dataframe_1_original.columns).all())
        self.assertTrue(dataframe_2.equals(dataframe_2_original))
        self.assertTrue((dataframe_2.index == dataframe_2_original.index).all())
        self.assertTrue((dataframe_2.columns == dataframe_2_original.columns).all())

        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1]))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1, dataframe_1]))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_1.copy()]))

        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.round(6)]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.round(5)]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1, dataframe_1.round(5)]))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.round(5)],
                                            float_tolerance=5))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.round(6),
                                                        dataframe_1.round(5)],
                                            float_tolerance=5))

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[0, 0]))
        dataframe_2.iat[0, 0] = dataframe_2.iat[0, 0] - 0.000001
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[1, 1]))
        dataframe_2.iat[1, 1] = 'c'
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[2, 2]))
        dataframe_2.iat[2, 2] = hs.RoundTo.MILLIONS
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[2, 3]))
        dataframe_2.iat[2, 3] = dataframe_2.iat[2, 3] + timedelta(seconds=1)
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[2, 3]))
        dataframe_2.iat[2, 3] = dataframe_2.iat[2, 3] + timedelta(seconds=1) - timedelta(seconds=1)
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[0, 0]))
        dataframe_2.iat[0, 0] = np.NaN
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[1, 1]))
        dataframe_2.iat[1, 1] = np.NaN
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[1, 2]))
        dataframe_2.iat[1, 2] = np.NaN
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))

        dataframe_2 = dataframe_1.copy()
        self.assertFalse(pd.isna(dataframe_2.iat[1, 3]))
        dataframe_2.iat[1, 3] = np.NaN
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_1.copy(), dataframe_2]))

        dataframe_2 = dataframe_1.copy()
        dataframe_2.index = ['a', 'b', 'c', 'd']
        dataframe_2.columns = ['a', 'b', 'c', 'd', 'e']
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2]))
        self.assertTrue(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2, dataframe_2]))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2], ignore_indexes=False))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2],
                                             ignore_column_names=False))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1],
                                             ignore_column_names=False))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_1, dataframe_2],
                                             ignore_indexes=False,
                                             ignore_column_names=False))
        self.assertFalse(hv.dataframes_match(dataframes=[dataframe_2, dataframe_1],
                                             ignore_indexes=False,
                                             ignore_column_names=False))

    def test_assert_dataframes_match(self):
        dataframe_1 = pd.DataFrame({'col': [1.123456789, 2.123456789, 3.123456789]})

        # test assertion errors when passing in invalid data
        with self.assertRaises(he.HelpskParamTypeError):
            hv.assert_dataframes_match(dataframes=None)  # noqa

        with self.assertRaises(he.HelpskParamTypeError):
            hv.assert_dataframes_match(dataframes=dataframe_1)  # noqa

        with self.assertRaises(he.HelpskParamValueError):
            hv.assert_dataframes_match(dataframes=[])

        with self.assertRaises(he.HelpskParamValueError):
            hv.assert_dataframes_match(dataframes=[dataframe_1])

        self.assertTrue(hv.assert_dataframes_match(dataframes=[dataframe_1, dataframe_1]) is None)

        dataframe_2 = dataframe_1.copy()
        dataframe_2.iat[0, 0] = np.nan

        with self.assertRaises(he.HelpskAssertionError):
            hv.assert_dataframes_match(dataframes=[dataframe_1, dataframe_2])

    def test_is_close(self):
        with self.assertRaises(Exception):
            hv.assert_is_close(np.nan, 1)
        with self.assertRaises(Exception):
            hv.assert_is_close(1, np.nan)
        with self.assertRaises(Exception):
            hv.assert_is_close(np.nan, np.nan)
        with self.assertRaises(Exception):
            hv.assert_is_close(None, 1)  # noqa
        with self.assertRaises(Exception):
            hv.assert_is_close(1, None)  # noqa
        with self.assertRaises(Exception):
            hv.assert_is_close(None, None)  # noqa

        def sub_test(value, tolerance, failure_tolerance):
            hv.assert_is_close(value, value)
            hv.assert_is_close(value, value + tolerance)
            hv.assert_is_close(value + tolerance, value)
            hv.assert_is_close(-value, -value - tolerance)
            hv.assert_is_close(-value - tolerance, -value)

            with self.assertRaises(he.HelpskAssertionError):
                hv.assert_is_close(value, value + failure_tolerance)
            with self.assertRaises(he.HelpskAssertionError):
                hv.assert_is_close(value + failure_tolerance, value)
            with self.assertRaises(he.HelpskAssertionError):
                hv.assert_is_close(-value, -value - failure_tolerance)
            with self.assertRaises(he.HelpskAssertionError):
                hv.assert_is_close(-value + -failure_tolerance, -value)

        sub_test(value=0, tolerance=0.000001, failure_tolerance=0.0000011)
        sub_test(value=1, tolerance=0.000001, failure_tolerance=0.0000011)
        sub_test(value=100, tolerance=0.000001, failure_tolerance=0.0000011)
        sub_test(value=-100, tolerance=0.000001, failure_tolerance=0.0000011)

    def test_raises_exception(self):

        def my_function_exception():
            raise ValueError()

        def my_function_runs():
            return True

        self.assertTrue(hv.raises_exception(my_function_exception))
        # should return True since my_function_exception raises ValueError
        self.assertTrue(hv.raises_exception(my_function_exception, ValueError))
        # should return False since my_function_exception raises ValueError, not TypeError
        self.assertFalse(hv.raises_exception(my_function_exception, TypeError))
        self.assertFalse(hv.raises_exception(my_function_runs))


if __name__ == '__main__':
    unittest.main()



"""Contains the base/abstract classes that wraps the connection/querying logic of various databases.

See documentation in database.py
"""
from abc import ABCMeta, abstractmethod
import time
import pandas as pd
from helpsk.utility import suppress_stdout, suppress_warnings


class Configuration(metaclass=ABCMeta):
    """A basic configuration object will product a dictionary representing the keyword parameters that can
    be passed to a Database object.
    """
    @abstractmethod
    def get_dictionary(self) -> dict:
        """
        Returns:
             a dictionary to be passed to the Database object
        """


class Database(metaclass=ABCMeta):
    """
    Base class that wraps the connection/querying logic of various databases.
    """
    def __init__(self, **kwargs):
        self.connection_object = None

    @classmethod
    def from_config(cls, config: Configuration):
        """Takes a Configuration object that contains the connection details to the database.
        """
        return cls(**config.get_dictionary())

    def is_connected(self):
        """
        Returns:
            True if the database is connected, otherwise False
        """
        return self.connection_object is not None

    @abstractmethod
    def _open_connection_object(self) -> object:
        """Child classes will implement the logic to connect to the database and return the connection object.

        The returning value will be stored in self.connection_object

        Returns:
            the connection object (usually returned from the underlying `connect()` method. For example, the
            returning value from `snowflake.connection.connect()`
        """

    @abstractmethod
    def _close_connection_object(self):
        """Child classes will implement the logic to close the connection to the database, typically by
        calling `close()` on the connection object returned from _open_connection_object.

        For example:
            self.connection_object.close()
        """

    def connect(self):
        """Call this method to open the connection to the database.

        Alternatively you can use the context manager:

        ```
        with Database(...) as database:
            database.query("SELECT * FROM table LIMIT 100")
        ```
        """
        if not self.is_connected():
            self.connection_object = self._open_connection_object()
            assert self.is_connected()

    def close(self):
        """Call this method to close the connection to the database.

        Alternatively you can use the context manager:

        ```
        with Database(...) as database:
            database.query("SELECT * FROM table LIMIT 100")
        ```
        """
        if self.is_connected():
            self._close_connection_object()
            self.connection_object = None
            assert not self.is_connected()

    def __enter__(self):
        """Enables the context manager"""
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Enables the context manager"""
        self.close()

    @abstractmethod
    def _query(self, sql: str) -> pd.DataFrame:
        """
        Method for child classes to override that contains logic to query

        Args:
            sql:
                SQL to execute e.g. "SELECT * FROM table LIMIT 100"

        Returns:
            a pandas Dataframe with the results from the query
        """

    def query(self, sql: str, *,
              show_messages: bool = False, show_warnings: bool = False,
              show_elapsed_time: bool = True) -> pd.DataFrame:
        """Queries the database and returns the results as a pandas Dataframe

        Args:
            sql:
                SQL to execute e.g. "SELECT * FROM table LIMIT 100"
            show_messages:
                if True, shows messages generated from the underlying query
            show_warnings:
                if True, shows warnings generated from the underlying query
            show_elapsed_time:
                if True, shows the elapsed execution time for the query

        Returns:
            a pandas Dataframe with the results from the query
        """
        assert self.is_connected()
        start_time = time.time()
        if show_messages and show_warnings:
            with suppress_stdout(), suppress_warnings():
                results = self._query(sql=sql)
        elif show_messages:
            with suppress_stdout():
                results = self._query(sql=sql)
        elif show_warnings:
            with suppress_warnings():
                results = self._query(sql=sql)
        else:
            results = self._query(sql=sql)

        end_time = time.time()

        if show_elapsed_time:
            elapsed_time = end_time - start_time
            time_units = 'minutes' if elapsed_time > 60 else 'seconds'
            elapsed_time = elapsed_time / 60 if elapsed_time > 60 else elapsed_time
            elapsed_time = round(elapsed_time, 1)
            print(f'Elapsed Time: {elapsed_time} {time_units}')

        return results

    @abstractmethod
    def execute_statement(self, statement: str):
        """This method executes a statement without any data returned."""

    @abstractmethod
    def insert_records(self,
                        dataframe: pd.DataFrame,
                        table: str,
                        create_table: bool = False,
                        overwrite: bool = True,
                        schema: str = None,
                        database: str = None):
        """
        This method inserts rows into a table from a pandas DataFrame.

        Args:
            dataframe:
                the pandas dataframe to insert
            table:
                the name of the table
            create_table:
                if True, creates the table before inserting
            overwrite:
                if True, drops all records before inserting
            schema:
                the name of the schema
            database:
                the name of the database
        """



"""Contains a collection of helper functions to manipulate strings.
"""
from enum import unique, Enum, auto
from typing import List, Union


def collapse(*args: Union[str, List[str]], separate: str = '', surround: str = '') -> str:
    """Takes a list of strings and concatenates them, separating each string with the value of `separate` and
    surrounding each string with the value of `surround`.

    Args:
        *args:
            list of strings to concatenate
        separate:
            string used to separate each string (default '')
        surround:
            string used to surround each string (default '')

    Returns:
        string surrounded by `surround` and separated by `separate`
    """
    if len(args) == 1 and isinstance(args[0], list):
        args = args[0]

    return separate.join([surround + x + surround for x in args])


@unique
class RoundTo(Enum):
    """Valid values for rounding numbers; used in e.g. `format_number()`
    """
    NONE = auto()
    THOUSANDS = 'K'
    MILLIONS = 'M'
    BILLIONS = 'B'
    TRILLIONS = 'T'
    AUTO = auto()


def format_number(value: float,
                  granularity: RoundTo = RoundTo.AUTO,
                  places: int = 2) -> str:
    """Formats numbers in a way that humans would typically want to see at a glance (e.g. on a graph.)

    For example, `123560000` is transformed to `123.56M`

    Args:
        value : the number (float) to format

        granularity : the granularity (e.g. thousands, millions, billions). (default is
            RoundTo.AUTO)

            `RoundTo.NONE` will not change the granularity, but will still add commas if necessary.

            `RoundTo.AUTO` will reduce the granularity (or not) depending on the number passed in.

        places: the number of digits to the right of the decimal to display. (default is `2`)

            For example:
                `format_number(123567000, granularity=RoundTo.MILLIONS, num_digits=2)` -> '123.57M'
                `format_number(123567000, granularity=RoundTo.MILLIONS, num_digits=3)` -> '123.567M'

    Returns:
        the formatted number as a string
    """
    granularity_lookup = {RoundTo.THOUSANDS: 1000,
                          RoundTo.MILLIONS: 1000000,
                          RoundTo.BILLIONS: 1000000000,
                          RoundTo.TRILLIONS: 1000000000000}

    if granularity == RoundTo.NONE:
        granularity = ''
    elif granularity == RoundTo.AUTO:

        if abs(value) >= granularity_lookup[RoundTo.TRILLIONS]:
            value = value / granularity_lookup[RoundTo.TRILLIONS]
            granularity = RoundTo.TRILLIONS.value

        elif abs(value) >= granularity_lookup[RoundTo.BILLIONS]:
            value = value / granularity_lookup[RoundTo.BILLIONS]
            granularity = RoundTo.BILLIONS.value

        elif abs(value) >= granularity_lookup[RoundTo.MILLIONS]:
            value = value / granularity_lookup[RoundTo.MILLIONS]
            granularity = RoundTo.MILLIONS.value

        elif abs(value) >= granularity_lookup[RoundTo.THOUSANDS]:
            value = value / granularity_lookup[RoundTo.THOUSANDS]
            granularity = RoundTo.THOUSANDS.value

        elif abs(value) >= 1:
            granularity = ''
        else:
            # else we have a number that is less than one, so we will override places
            granularity = ''

            places = 1
            while abs(value) < pow(10, places * -1):
                places += 1

            places += 2
    else:
        value = value / granularity_lookup[granularity]
        granularity = granularity.value

    assert isinstance(granularity, str)

    # return '{{:,.{0}f}}'.format(places).format(value) + granularity
    return f'{value:,.{places}f}' + granularity



"""A collection of system and environment related helper functions.
"""
import inspect
import warnings
import datetime
import sys
import os
import pickle
from pathlib import Path
import pandas as pd

from yaml import safe_load
from contextlib import contextmanager, redirect_stdout


def open_yaml(file_path):
    """Open a yaml file via yaml.safe_load."""
    with open(file_path, "r") as f:
        return safe_load(f)


def read_pickle(path):
    """
    Simple helper function that read's from a pickled object.

    Args:
        path:
            File path where the pickled object will be stored.
    """
    with open(path, 'rb') as handle:
        unpickled_object = pickle.load(handle)
    return unpickled_object


def to_pickle(obj, path):
    """
    Simple helper function that saves a pickled object.

    Args:
        obj:
            the object to save
        path:
            File path where the pickled object will be read from.
    """
    with open(path, 'wb') as handle:
        pickle.dump(obj, handle)


def dataframe_to_pickle(df: pd.DataFrame, output_directory: str, file_name: str) -> str:
    """
    This function takes a Pandas DataFrame and saves it as a pickled object to the directory with the file
    name specified. The output directory is created if it does not yet exist.

    Args:
        df: the Pandas DataFrame to pickle
        output_directory: the directory to save the pickled object
        file_name: the name of the file
    """
    Path(output_directory).mkdir(exist_ok=True)
    file_path = os.path.join(output_directory, file_name)
    df.to_pickle(file_path)
    return file_path


def dataframe_to_csv(df: pd.DataFrame, output_directory: str, file_name: str) -> str:
    """
    This function takes a Pandas DataFrame and saves it as a csv file to the directory with the file
    name specified. The output directory is created if it does not yet exist.

    Args:
        df: the Pandas DataFrame to pickle
        output_directory: the directory to save the csv file
        file_name: the name of the file
    """
    Path(output_directory).mkdir(exist_ok=True)
    file_path = os.path.join(output_directory, file_name)
    df.to_csv(file_path, index=False)
    return file_path


def object_to_pickle(obj: object, output_directory: str, file_name: str) -> str:
    """
    This function takes a generic object and saves it as a pickled object to the directory with the file
    name specified. The output directory is created if it does not yet exist.

    Args:
        obj: the object to pickle
        output_directory: the directory to save the pickled object
        file_name: the name of the file
    """
    Path(output_directory).mkdir(exist_ok=True)
    file_path = os.path.join(output_directory, file_name)
    to_pickle(obj=obj, path=file_path)
    return file_path


@contextmanager
def redirect_stdout_to_file(file, mode='w'):
    """ Helper context manager that opens a file and redirects standard output to that file.

    Example:

    ```
    with redirect_stdout_to_file(file_name):
        print_dataframe(dataframe)
    ```

    Args:
        file:
            the name and path of the file to open (argument is passed to `open()`)
        mode:
            the mode of the file e.g. `w` (argument is passed to `open()`)
    """
    with open(file, mode) as handle:
        with redirect_stdout(handle):
            yield


@contextmanager
def suppress_stdout():
    """Suppress Output


    ```
    print("Now you see it")
    with suppress_stdout():
        print("Now you don't")
    ```

    code from: https://stackoverflow.com/questions/2125702/how-to-suppress-console-output-in-python
    """
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


@contextmanager
def suppress_warnings():
    """Simple Wrapper around warnings.catch_warnings()"""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        yield


def is_debugging():
    """
    Returns:
        Returns True if the environment is currently debugging (e.g. set a breakpoint in PyCharm),
        False if not
    """
    for frame in inspect.stack():
        if frame[1].endswith("pydevd.py"):
            return True

    return False


class Timer:
    """
    This class provides way to time the duration of code within the context manager.
    """
    def __init__(self, message):
        self._message = message

    def __enter__(self):
        print(f'Timer Started: {self._message}')
        self._start = datetime.datetime.now()
        return self

    def __exit__(self, *args):
        self._end = datetime.datetime.now()
        self._interval = self._end - self._start
        print(f'Timer Finished ({self._interval.total_seconds():.2f} seconds)')
